# encoding: utf-8

import datetime
import decimal
import logging
import random
import re
import uuid
from string import ascii_letters as letters
from unittest import (
    skip,
    skipIf,
)

# LIBRARIES
import sleuth
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import (
    FieldError,
    ValidationError,
)
from django.db import (
    DatabaseError,
    DataError,
    IntegrityError,
    NotSupportedError,
)
from django.db import connection
from django.db import connection as default_connection
from django.db import models
from django.db.models.query import Q
from django.forms import ModelForm
from django.forms.models import modelformset_factory
from django.test import RequestFactory
from django.test.utils import override_settings
from django.urls import path
from django.utils.safestring import SafeText

from django.utils.timezone import make_aware
from google.cloud.datastore import key
from google.cloud.datastore.entity import Entity

from gcloudc.db.backends.datastore import (
    caching,
    indexing,
    transaction,
)

from gcloudc.db.backends.datastore.expressions import Scatter

from gcloudc.db.backends.datastore.commands import FlushCommand
from gcloudc.db.backends.datastore.indexing import (
    IExactIndexer,
    add_special_index,
    get_indexer,
)
from gcloudc.db.backends.datastore.unique_mixins import UniquenessMixin
from gcloudc.db.backends.datastore.unique_utils import (
    _unique_combinations,
    query_is_unique,
    unique_identifiers_from_entity,
)
from gcloudc.db.backends.datastore.utils import (
    decimal_to_string,
    entity_matches_query,
    normalise_field_value,
)
from gcloudc.db.decorators import disable_cache

from . import TestCase
from .models import (
    Animal,
    DateTimeModel,
    DurationModel,
    Enclosure,
    IntegerModel,
    ModelWithDates,
    ModelWithNullableCharField,
    ModelWithUniques,
    ModelWithUniquesOnForeignKey,
    MultiTableChildOne,
    MultiTableChildTwo,
    MultiTableParent,
    NullDate,
    NullDateSet,
    Permission,
    SelfRelatedModel,
    SpecialIndexesModel,
    TestFruit,
    TestUser,
    UniqueModel,
    UniqueModelWithLongPK,
    UUIDTestModel,
    Zoo,
)

DEFAULT_NAMESPACE = default_connection.ops.connection.settings_dict.get("NAMESPACE")


class BackendTests(TestCase):
    def test_pk_gt_empty_returns_all(self):
        for i in range(10):
            TestFruit.objects.create(name=str(i), color=str(i))

        self.assertEqual(10, TestFruit.objects.filter(pk__gt="").count())
        self.assertEqual(10, TestFruit.objects.filter(pk__gte="").count())
        self.assertEqual(0, TestFruit.objects.filter(pk__lt="").count())
        self.assertEqual(0, TestFruit.objects.filter(pk__lte="").count())

    def test_pk_gt_zero_returns_all(self):
        IntegerModel.objects.create(pk=1, integer_field=1)
        IntegerModel.objects.create(pk=2, integer_field=2)

        results = IntegerModel.objects.filter(pk__gt=0)
        self.assertEqual(2, len(results))

        results = IntegerModel.objects.filter(pk__gte=0)
        self.assertEqual(2, len(results))

        results = IntegerModel.objects.filter(pk__lt=0)
        self.assertEqual(0, len(results))

        results = IntegerModel.objects.filter(pk__lte=0)
        self.assertEqual(0, len(results))

    def test_entity_matches_query(self):
        rpc = transaction._rpc(default_connection.alias)

        entity = Entity(rpc.key("test_model"))
        entity["name"] = "Charlie"
        entity["age"] = 22

        query = rpc.query(kind="test_model")
        query.add_filter("name", "=", "Charlie")
        self.assertTrue(entity_matches_query(entity, query))

        query = rpc.query(kind="test_model")
        query.add_filter("age", ">=", 5)
        self.assertTrue(entity_matches_query(entity, query))

        query = rpc.query(kind="test_model")
        query.add_filter("age", "<", 22)
        self.assertFalse(entity_matches_query(entity, query))

        query = rpc.query(kind="test_model")
        query.add_filter("age", "<=", 22)
        self.assertTrue(entity_matches_query(entity, query))

        query = rpc.query(kind="test_model")
        query.add_filter("name", "=", "Fred")
        self.assertFalse(entity_matches_query(entity, query))

        # If the entity has a list field, then if any of them match the
        # query then it's a match
        entity["name"] = ["Bob", "Fred", "Dave"]
        self.assertTrue(entity_matches_query(entity, query))  # ListField test

    def test_exclude_pks_with_slice(self):
        for i in range(10):
            TestFruit.objects.create(name=str(i), color=str(i))

        to_exclude = [str(x) for x in list(range(5)) + list(range(15, 20))]

        to_return = TestFruit.objects.exclude(pk__in=set(to_exclude)).values_list("pk", flat=True)[:2]
        self.assertEqual(2, len(to_return))

        qs = TestFruit.objects.filter(
            pk__in=to_return
        )

        self.assertEqual(2, len(qs))

    def test_count_on_excluded_pks(self):
        TestFruit.objects.create(name="Apple", color="Red")
        TestFruit.objects.create(name="Orange", color="Orange")

        self.assertEqual(1, TestFruit.objects.filter(pk__in=["Apple", "Orange"]).exclude(pk__in=["Apple"]).count())

    def test_defaults(self):
        fruit = TestFruit.objects.create(name="Apple", color="Red")
        self.assertEqual("Unknown", fruit.origin)

        rpc = transaction._rpc(default_connection.alias)

        instance = rpc.get(rpc.key(TestFruit._meta.db_table, fruit.pk))
        del instance["origin"]
        rpc.put(instance)

        fruit = TestFruit.objects.get()
        self.assertIsNone(fruit.origin)
        fruit.save()

        fruit = TestFruit.objects.get()
        self.assertEqual("Unknown", fruit.origin)

    @disable_cache()
    def test_get_by_keys(self):
        colors = ["Red", "Green", "Blue", "Yellow", "Orange"]
        fruits = [TestFruit.objects.create(name=str(x), color=random.choice(colors)) for x in range(32)]
        rpc = transaction._rpc(default_connection.alias)

        # Check that projections work with key lookups
        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.query") as query:
            TestFruit.objects.only("color").get(pk="0").color
            self.assertEqual(query.calls[0].kwargs["projection"], ["color"])

            # Make sure the query is an ancestor of the key
            self.assertEqual(
                query.call_returns[0].ancestor,
                rpc.key(
                    TestFruit._meta.db_table, "0"
                )
            )

        # Now check projections work with fewer than 100 things
        with sleuth.watch('gcloudc.db.backends.datastore.meta_queries.AsyncMultiQuery.__init__') as query_init:
            with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.query") as query:
                keys = [str(x) for x in range(32)]
                results = list(TestFruit.objects.only("color").filter(pk__in=keys).order_by("name"))  # One multiquery
                self.assertEqual(query_init.call_count, 1)
                self.assertEqual(len([x.ancestor for x in query.call_returns if x.ancestor]), 32)  # 32 Ancestor calls

                # Confirm the ordering is correct
                self.assertEqual(sorted(keys), [x.pk for x in results])

        results = list(TestFruit.objects.only("color").filter(pk__in=keys).order_by("name")[5:10])
        self.assertEqual(len(results), 5)
        self.assertEqual([x.pk for x in results], sorted(keys)[5:10])

        # Make sure we can do a normal (non-projection) get by keys
        self.assertItemsEqual(TestFruit.objects.filter(pk__in=keys), fruits)

    def test_get_or_create(self):
        """
            Django's get_or_create can do the following:

            1. get(**lookup) -> throws DoesNotExist
            2. Catches DoesNotExist
            3. create() -> throws IntegrityError
            4. get(**lookup)

            This test proves that we throw the right kind of error at step 3 when
            unique constraints are violated.
        """

        def wrap_get(func):
            def _wrapped(*args, **kwargs):
                try:
                    if _wrapped.calls == 0:
                        raise UniqueModel.DoesNotExist()
                    else:
                        return func(*args, **kwargs)
                finally:
                    _wrapped.calls += 1

            _wrapped.calls = 0
            return _wrapped

        from django.db.models import query
        wrapped_get = wrap_get(query.QuerySet.get)

        UniqueModel.objects.create(unique_field="Test")

        with disable_cache():
            with sleuth.switch("django.db.models.query.QuerySet.get", wrapped_get):
                instance, created = UniqueModel.objects.get_or_create(unique_field="Test")
                self.assertFalse(created)

    def test_setting_non_null_null_throws_integrity_error(self):
        with self.assertRaises(IntegrityError):
            IntegerModel.objects.create(integer_field=None)

        with self.assertRaises(IntegrityError):
            instance = IntegerModel()
            instance.integer_field = None
            instance.save()

        with self.assertRaises(IntegrityError):
            instance = IntegerModel.objects.create(integer_field=1)
            instance = IntegerModel.objects.get()
            instance.integer_field = None
            instance.save()

    def test_normalise_field_value(self):
        self.assertEqual(u'0000475231073257', normalise_field_value(decimal.Decimal(475231073257)))
        self.assertEqual(u'-0000475231073257', normalise_field_value(decimal.Decimal(-475231073257)))
        self.assertEqual(u'0000000004752311', normalise_field_value(decimal.Decimal(4752310.73257)))
        self.assertEqual(u'0000004752310733', normalise_field_value(decimal.Decimal(4752310732.57)))
        self.assertEqual(
            datetime.datetime(2015, 1, 27, 2, 46, 8, 584258),
            normalise_field_value(datetime.datetime(2015, 1, 27, 2, 46, 8, 584258))
        )

    def test_decimal_to_string(self):
        self.assertEqual(u'0002312487812767', decimal_to_string(decimal.Decimal(2312487812767)))
        self.assertEqual(u'-0002312487812767', decimal_to_string(decimal.Decimal(-2312487812767)))
        self.assertEqual(u'002312487812', decimal_to_string(decimal.Decimal(2312487812), 12))
        self.assertEqual(u'002387812.320', decimal_to_string(decimal.Decimal(2387812.32), 12, 3))
        self.assertEqual(u'-002387812.513', decimal_to_string(decimal.Decimal(-2387812.513212), 12, 3))
        self.assertEqual(u'0237812.000', decimal_to_string(decimal.Decimal(237812), 10, 3))
        self.assertEqual(u'-0237812.210', decimal_to_string(decimal.Decimal(-237812.21), 10, 3))

    def test_gae_conversion(self):
        # A PK IN query should result in a single get by key

        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.get") as get_mock:
            list(TestUser.objects.filter(pk__in=[1, 2, 3]))  # Force the query to run
            self.assertEqual(1, get_mock.call_count)

        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.query") as query_mock:
            list(TestUser.objects.filter(username="test"))
            self.assertEqual(1, query_mock.call_count)

        with sleuth.watch("gcloudc.db.backends.datastore.meta_queries.AsyncMultiQuery.fetch") as query_mock:
            list(TestUser.objects.filter(username__in=["test", "cheese"]))
            self.assertEqual(1, query_mock.call_count)

        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.get") as get_mock:
            list(TestUser.objects.filter(pk=1))
            self.assertEqual(1, get_mock.call_count)

        with sleuth.watch("gcloudc.db.backends.datastore.meta_queries.AsyncMultiQuery.fetch") as query_mock:
            list(TestUser.objects.exclude(username__startswith="test"))
            self.assertEqual(1, query_mock.call_count)

        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.get") as get_mock:
            list(
                TestUser.objects.filter(pk__in=[1, 2, 3, 4, 5, 6, 7, 8]).
                filter(username__in=["test", "test2", "test3"]).
                filter(email__in=["test@example.com", "test2@example.com"])
            )

            self.assertEqual(1, get_mock.call_count)

    def test_gae_query_display(self):
        # Shouldn't raise any exceptions:
        representation = str(TestUser.objects.filter(username='test').query)
        self.assertTrue('test' in representation)
        self.assertTrue('username' in representation)

    def test_range_behaviour(self):
        IntegerModel.objects.create(integer_field=5)
        IntegerModel.objects.create(integer_field=10)
        IntegerModel.objects.create(integer_field=15)

        self.assertItemsEqual(
            [10],
            IntegerModel.objects.filter(
                integer_field__range=(6, 14)
            ).values_list("integer_field", flat=True)
        )

        self.assertItemsEqual(
            [5, 10, 15],
            IntegerModel.objects.filter(
                integer_field__range=(5, 15)
            ).order_by("integer_field").values_list("integer_field", flat=True)
        )

        self.assertItemsEqual(
            [5, 15],
            IntegerModel.objects.exclude(
                integer_field__range=(6, 14)
            ).values_list("integer_field", flat=True)
        )

    def test_exclude_nullable_field(self):
        instance = ModelWithNullableCharField.objects.create(some_id=999)  # Create a nullable thing
        ModelWithNullableCharField.objects.create(some_id=999, field1="test")  # Create a nullable thing
        self.assertItemsEqual(
            [instance],
            ModelWithNullableCharField.objects.filter(
                some_id=999
            ).exclude(field1="test").all()
        )

        instance.field1 = "bananas"
        instance.save()

        self.assertEqual(instance, ModelWithNullableCharField.objects.filter(some_id=999).exclude(field1="test")[0])

    def test_null_date_field(self):
        null_date = NullDate()
        null_date.save()

        null_date = NullDate.objects.get()
        self.assertIsNone(null_date.date)
        self.assertIsNone(null_date.time)
        self.assertIsNone(null_date.datetime)

    def test_convert_str_subclasses_to_str(self):
        # The App Engine SDK raises BadValueError if you try saving a SafeText
        # string to a CharField. Djangae explicitly converts it to str.
        grue = SafeText(u'grue')

        self.assertIsInstance(grue, str)
        self.assertNotEqual(type(grue), str)

        obj = TestFruit.objects.create(name=u'foo', color=grue)
        obj = TestFruit.objects.get(pk=obj.pk)
        self.assertEqual(type(obj.color), str)

        obj = TestFruit.objects.filter(color=grue)[0]
        self.assertEqual(type(obj.color), str)

    def test_notsupportederror_thrown_on_too_many_inequalities(self):
        TestFruit.objects.create(name="Apple", color="Green", origin="England")
        pear = TestFruit.objects.create(name="Pear", color="Green")
        banana = TestFruit.objects.create(name="Banana", color="Yellow")

        # Excluding one field is fine
        self.assertItemsEqual([pear, banana], list(TestFruit.objects.exclude(name="Apple")))

        # Excluding a field, and doing a > or < on another is not so fine
        with self.assertRaises(NotSupportedError):
            self.assertEqual(pear, TestFruit.objects.exclude(origin="England").filter(color__lt="Yellow").get())

        # Same with excluding two fields
        with self.assertRaises(NotSupportedError):
            list(TestFruit.objects.exclude(origin="England").exclude(color="Yellow"))

        # But apparently excluding the same field twice is OK
        self.assertItemsEqual(
            [banana],
            list(TestFruit.objects.exclude(
                origin="England"
            ).exclude(name="Pear").order_by("origin"))
        )

        # And apparently having both a __gt and a __lt filter on the same field is also fine
        self.assertItemsEqual([banana], list(TestFruit.objects.order_by().filter(name__lt="Pear", name__gt="Apple")))

    def test_excluding_pks_is_emulated(self):
        apple = TestFruit.objects.create(name="Apple", color="Green", is_mouldy=True, origin="England")
        banana = TestFruit.objects.create(name="Banana", color="Yellow", is_mouldy=True, origin="Dominican Republic")
        cherry = TestFruit.objects.create(name="Cherry", color="Red", is_mouldy=True, origin="Germany")
        pear = TestFruit.objects.create(name="Pear", color="Green", origin="England")

        self.assertEqual(
            [apple, pear],
            list(TestFruit.objects.filter(origin__lt="Germany").exclude(
                pk=banana.pk
            ).exclude(pk=cherry.pk).order_by("origin"))
        )

        self.assertEqual(
            [apple, cherry],
            list(TestFruit.objects.exclude(
                origin="Dominican Republic"
            ).exclude(pk=pear.pk).order_by("origin"))
        )

        self.assertEqual(
            [],
            list(TestFruit.objects.filter(
                is_mouldy=True
            ).filter(
                color="Green", origin__gt="England"
            ).exclude(pk=pear.pk).order_by("-origin"))
        )

        self.assertEqual([cherry, banana], list(TestFruit.objects.exclude(pk=pear.pk).order_by("-name")[:2]))
        self.assertEqual([banana, apple], list(TestFruit.objects.exclude(pk=pear.pk).order_by("origin", "name")[:2]))

    def test_datetime_fields(self):
        date = datetime.datetime.today()
        dt = datetime.datetime.now()
        time = datetime.time(0, 0, 0)

        # check if creating objects work
        obj = NullDate.objects.create(date=date, datetime=dt, time=time)

        # check if filtering objects work
        self.assertItemsEqual([obj], NullDate.objects.filter(datetime=dt))
        self.assertItemsEqual([obj], NullDate.objects.filter(date=date))
        self.assertItemsEqual([obj], NullDate.objects.filter(time=time))

        # check if updating objects work
        obj.date = date + datetime.timedelta(days=1)
        obj.datetime = dt + datetime.timedelta(days=1)
        obj.time = datetime.time(23, 0, 0)
        obj.save()
        self.assertItemsEqual([obj], NullDate.objects.filter(datetime=obj.datetime))
        self.assertItemsEqual([obj], NullDate.objects.filter(date=obj.date))
        self.assertItemsEqual([obj], NullDate.objects.filter(time=obj.time))

    def test_related_datetime_nullable(self):
        date = datetime.datetime.today()
        dt = datetime.datetime.now()
        time = datetime.time(0, 0, 0)

        date_set = NullDateSet.objects.create()
        empty_obj = NullDate.objects.create(date=None, datetime=None, time=None)
        date_set.dates.add(empty_obj)

        obj = NullDate.objects.create(date=date, datetime=dt, time=time)
        date_set.dates.add(obj)
        date_set.save()

        # check if filtering/excluding of None works in RelatedSetField
        self.assertItemsEqual([obj], date_set.dates.filter(datetime__isnull=False))
        self.assertItemsEqual([obj], date_set.dates.filter(date__isnull=False))
        self.assertItemsEqual([obj], date_set.dates.filter(time__isnull=False))

        self.assertItemsEqual([obj], date_set.dates.exclude(datetime=None))
        self.assertItemsEqual([obj], date_set.dates.exclude(date=None))
        self.assertItemsEqual([obj], date_set.dates.exclude(time=None))

        # sorting should work too
        self.assertItemsEqual([obj, empty_obj], date_set.dates.order_by('datetime'))
        self.assertItemsEqual([empty_obj, obj], date_set.dates.order_by('-datetime'))
        self.assertItemsEqual([obj, empty_obj], date_set.dates.order_by('date'))
        self.assertItemsEqual([empty_obj, obj], date_set.dates.order_by('-date'))
        self.assertItemsEqual([obj, empty_obj], date_set.dates.order_by('time'))
        self.assertItemsEqual([empty_obj, obj], date_set.dates.order_by('-time'))

    def test_update_with_f_expr(self):
        i = IntegerModel.objects.create(integer_field=1000)
        qs = IntegerModel.objects.all()
        qs.update(integer_field=models.F('integer_field') + 1)

        self.assertRaises(IntegerModel.DoesNotExist, IntegerModel.objects.get, integer_field=1000)
        i = IntegerModel.objects.get(pk=i.pk)
        self.assertEqual(1001, i.integer_field)

    def test_save_with_f_expr(self):
        i = IntegerModel.objects.create(integer_field=1000)

        i.integer_field = models.F('integer_field') + 1
        i.save()

        self.assertRaises(IntegerModel.DoesNotExist, IntegerModel.objects.get, integer_field=1000)
        i = IntegerModel.objects.get(pk=i.pk)
        self.assertEqual(1001, i.integer_field)

    def test_ordering_by_scatter_property(self):
        try:
            list(TestFruit.objects.order_by(Scatter()))
        except:  # noqa
            raise
            logging.exception("Error sorting on __scatter__")
            self.fail("Unable to sort on __scatter__ property like we should")

    def test_ordering_on_non_indexed_fields_not_supported(self):
        self.assertRaises(NotSupportedError, list, TestFruit.objects.order_by("text_field"))
        self.assertRaises(NotSupportedError, list, TestFruit.objects.order_by("binary_field"))

    def test_ordering_on_sparse_field(self):
        """
        Case when sorting on field that is not present in all of
        Datastore entities. That can easily happen when you added
        new field to model and did not populated all existing entities
        """
        rpc = transaction._rpc(default_connection.alias)

        # Clean state
        self.assertEqual(TestFruit.objects.count(), 0)

        # Put constistent instances to Datastore
        TestFruit.objects.create(name='a', color='a')
        TestFruit.objects.create(name='b', color='b')

        # Put inconsistent instances to Datastore
        # Color fields is missing (not even None)
        # we need more than 1 so we explore all sorting branches
        values = {'name': 'c'}
        entity = Entity(rpc.key(TestFruit._meta.db_table, 'c'))
        entity.update(values)
        rpc.put(entity)

        values = {'name': 'd'}
        entity = Entity(rpc.key(TestFruit._meta.db_table, 'd'))
        entity.update(values)
        rpc.put(entity)

        # Ok, we can get all 4 instances
        self.assertEqual(TestFruit.objects.count(), 4)

        # Sorted list. No exception should be raised
        # (esp KeyError from django_ordering_comparison)
        with sleuth.watch('gcloudc.db.backends.datastore.meta_queries.django_ordering_comparison') as compare:
            all_names = ['a', 'b', 'c', 'd']
            fruits = list(
                TestFruit.objects.filter(name__in=all_names).order_by('color', 'name')
            )
            # Make sure troubled code got triggered
            # ie. with all() it doesn't
            self.assertTrue(compare.called)

        # Test the ordering of the results.  The ones with a color of None should come back first,
        # and of the ones with color=None, they should be ordered by name
        # Missing one (None) as first
        expected_fruits = [
            ('c', None), ('d', None), ('a', 'a'), ('b', 'b'),
        ]

        self.assertEqual(
            [(fruit.name, fruit.color) for fruit in fruits],
            expected_fruits,
        )

    def test_duration_field_stored_as_float(self):
        """ See issue #512.  We have a bug report that the DurationField comes back as None when
            the value is set to a particular value which is roughly 3 days. This is caused by it
            being stored as a float instead of an int in the DB.
        """
        td2 = datetime.timedelta(days=2)
        # If the duration value is stored as a float instead of an int then this particular duration
        # will cause django.db.backends.base.operations.BaseDatabaseOperations.convert_durationfield_value
        # to return the value as None
        td3 = datetime.timedelta(days=3, seconds=14658, microseconds=886540)
        durations_as_2 = DurationModel.objects.create(
            duration_field=td2,
            duration_field_nullable=td2
        )
        durations_as_3 = DurationModel.objects.create(
            duration_field=td3,
            duration_field_nullable=td3
        )
        self.assertEqual(durations_as_2.duration_field, td2)
        self.assertEqual(durations_as_2.duration_field_nullable, td2)
        self.assertEqual(durations_as_3.duration_field, td3)
        self.assertEqual(durations_as_3.duration_field_nullable, td3)
        durations_as_2 = DurationModel.objects.get(pk=durations_as_2.pk)
        durations_as_3 = DurationModel.objects.get(pk=durations_as_3.pk)
        self.assertEqual(durations_as_2.duration_field, td2)
        self.assertEqual(durations_as_2.duration_field_nullable, td2)
        self.assertEqual(durations_as_3.duration_field, td3)
        self.assertEqual(durations_as_3.duration_field_nullable, td3)
        # And just for good measure, check the raw value in the datastore
        rpc = transaction._rpc(default_connection.alias)

        key = rpc.key(DurationModel._meta.db_table, durations_as_3.pk)
        entity = rpc.get(key)
        self.assertTrue(isinstance(entity['duration_field'], int))

    def test_datetime_and_time_fields_precision_for_projection_queries(self):
        """
        The returned datetime and time values should include microseconds.
        See issue #707.
        """

        t = datetime.time(22, 13, 50, 541022)
        dt = make_aware(datetime.datetime(2016, 5, 27, 18, 40, 12, 927371))
        NullDate.objects.create(time=t, datetime=dt)
        result = list(NullDate.objects.all().values_list('time', 'datetime'))
        expected = [(t, dt)]

        self.assertItemsEqual(result, expected)

    @override_settings(USE_TZ=False)
    def test_datetime_and_time_fields_precision_for_projection_queries_no_tz(self):
        """
        The returned datetime and time values should include microseconds.
        See issue #707.
        """

        t = datetime.time(22, 13, 50, 541022)
        dt = datetime.datetime(2016, 5, 27, 18, 40, 12, 927371)
        NullDate.objects.create(time=t, datetime=dt)
        result = list(NullDate.objects.all().values_list('time', 'datetime'))
        expected = [(t, dt)]

        self.assertItemsEqual(result, expected)

    @override_settings(TIMEZONE="Europe/Rome")
    def test_datetime_and_time_fields_precision_for_projection_queries_other_tz(self):
        """
        The returned datetime and time values should include microseconds.
        See issue #707.
        """

        t = datetime.time(22, 13, 50, 541022)
        dt = make_aware(datetime.datetime(2016, 5, 27, 18, 40, 12, 927371))
        NullDate.objects.create(time=t, datetime=dt)
        result = list(NullDate.objects.all().values_list('time', 'datetime'))
        expected = [(t, dt)]

        self.assertItemsEqual(result, expected)

    def test_filter_with_empty_q(self):
        t1 = TestUser.objects.create(username='foo', field2='bar')
        condition = Q() | Q(username='foo')
        self.assertEqual(t1, TestUser.objects.filter(condition).first())

        condition = Q()
        self.assertEqual(t1, TestUser.objects.filter(condition).first())

    def test_only_defer_does_project(self):
        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.query") as query:
            with sleuth.watch("google.cloud.datastore.query.Query.keys_only") as keys_only:
                list(TestUser.objects.only("pk").all())
                self.assertFalse(query.calls[0].kwargs["projection"])
                self.assertEqual(1, keys_only.call_count)

        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.query") as query:
            with sleuth.watch("google.cloud.datastore.query.Query.keys_only") as keys_only:
                list(TestUser.objects.values("pk"))
                self.assertFalse(query.calls[0].kwargs["projection"])
                self.assertEqual(1, keys_only.call_count)

        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.query") as query:
            with sleuth.watch("google.cloud.datastore.query.Query.keys_only") as keys_only:
                list(TestUser.objects.only("username").all())
                self.assertEqual(0, keys_only.call_count)
                self.assertItemsEqual(query.calls[0].kwargs["projection"], ["username"])

        with sleuth.watch("gcloudc.db.backends.datastore.transaction.Transaction.query") as query:
            with sleuth.watch("google.cloud.datastore.query.Query.keys_only") as keys_only:
                list(TestUser.objects.defer("username").all())
                self.assertEqual(0, keys_only.call_count)
                self.assertTrue(query.calls[0].kwargs["projection"])
                self.assertFalse("username" in query.calls[0].kwargs["projection"])

    def test_chaining_none_filter(self):
        t1 = TestUser.objects.create()
        self.assertFalse(TestUser.objects.none().filter(pk=t1.pk))

    def test_values_list_returns_str_strings(self):
        TestUser.objects.create(username=u"Łukasz")
        self.assertEqual(str, type(TestUser.objects.get().username))
        self.assertEqual(str, type(TestUser.objects.values_list("username", flat=True)[0]))


class ModelFormsetTest(TestCase):
    def test_reproduce_index_error(self):
        class TestModelForm(ModelForm):
            class Meta:
                model = TestUser
                fields = ("username", "email", "field2")

        test_model = TestUser.objects.create(username='foo', field2='bar')
        TestModelFormSet = modelformset_factory(TestUser, form=TestModelForm, extra=0)
        TestModelFormSet(queryset=TestUser.objects.filter(pk=test_model.pk))

        data = {
            'form-INITIAL_FORMS': 0,
            'form-MAX_NUM_FORMS': 0,
            'form-TOTAL_FORMS': 0,
            'form-0-id': test_model.id,
            'form-0-field1': 'foo_1',
            'form-0-field2': 'bar_1',
        }
        factory = RequestFactory()
        request = factory.post('/', data=data)

        TestModelFormSet(request.POST, request.FILES)


urlpatterns = [
    path("/", lambda x: None)
]


class CacheTests(TestCase):

    def test_cache_set(self):
        cache.set('test?', 'yes!')
        self.assertEqual(cache.get('test?'), 'yes!')

    def test_cache_timeout(self):
        cache.set('test?', 'yes!', 1)
        import time
        time.sleep(1)
        self.assertEqual(cache.get('test?'), None)

    @override_settings(ROOT_URLCONF="gcloudc.tests.test_connector")
    def test_cache_cleared_on_request(self):
        user = TestUser.objects.create(username="test")

        user_key = key.Key(
            TestUser._meta.db_table,
            user.pk,
            namespace=connection.settings_dict["NAMESPACE"],
            project=connection.settings_dict["PROJECT"]
        )

        from_cache = caching.get_from_cache_by_key(
            user_key
        )

        self.assertIsNotNone(from_cache)
        self.client.get("/")

        from_cache = caching.get_from_cache_by_key(
            user_key
        )

        self.assertIsNone(from_cache)

    def test_disable_cache_decorator(self):
        user = TestUser.objects.create(username="randy", first_name="Randy")

        with transaction.atomic():
            with sleuth.watch('gcloudc.db.backends.datastore.context.CacheDict.get') as cachedict_get:
                user.first_name = "Łukasz"
                user.save()

                from_cache = TestUser.objects.get(username="randy")

                # Ok, cache works
                self.assertEqual(from_cache.first_name, "Łukasz")
                self.assertEqual(cachedict_get.call_count, 1)

                with disable_cache():
                    non_cached = TestUser.objects.get(username="randy")
                    # Result is not fetched from the cache
                    self.assertEqual(non_cached.first_name, "Randy")
                    self.assertEqual(cachedict_get.call_count, 1)

                from_cache = TestUser.objects.get(username="randy")

                # Previous GET didn't override the cache
                self.assertEqual(from_cache.first_name, "Łukasz")
                self.assertEqual(cachedict_get.call_count, 2)


def compare_markers(list1, list2):
    return (
        sorted([(x.key(), x.instance) for x in list1]) == sorted([(x.key(), x.instance) for x in list2])
    )


class ConstraintTests(TestCase):
    """
        Tests for unique constraint handling
    """

    def test_conflicting_insert_throws_integrity_error(self):
        ModelWithUniques.objects.create(name="One")

        with self.assertRaises(IntegrityError):
            ModelWithUniques.objects.create(name="One")

        # An insert with a specified ID enters a different code path
        # so we need to ensure it works
        ModelWithUniques.objects.create(id=555, name="Two")

        with self.assertRaises(IntegrityError):
            ModelWithUniques.objects.create(name="Two")

        # Make sure that bulk create works properly
        ModelWithUniques.objects.bulk_create([
            ModelWithUniques(name="Three"),
            ModelWithUniques(name="Four"),
            ModelWithUniques(name="Five"),
        ])

        with self.assertRaises(IntegrityError):
            ModelWithUniques.objects.create(name="Four")

    def test_integrity_error_message_correct(self):
        """ Check that the IntegrityError messages mentions the correct field(s). """

        # Create a conflict on `unique_field`
        obj1 = UniqueModel.objects.create(unique_field="One")
        try:
            UniqueModel.objects.create(unique_field="One", unique_combo_one=1)
        except IntegrityError as e:
            self.assertTrue("unique_field" in str(e))

        # Create a conflict on `unique_relation`
        UniqueModel.objects.create(unique_relation=obj1, unique_field="two", unique_combo_one=2)
        try:
            UniqueModel.objects.create(unique_relation=obj1, unique_field="three", unique_combo_one=3)
        except IntegrityError as e:
            self.assertTrue("unique_relation" in str(e))

        # Create a conflict on a unique_together combo`
        UniqueModel.objects.create(unique_field="four", unique_combo_one=4, unique_combo_two="five")
        try:
            UniqueModel.objects.create(unique_field="five", unique_combo_one=4, unique_combo_two="five")
        except IntegrityError as e:
            self.assertTrue("unique_combo_one" in str(e))
            self.assertTrue("unique_combo_two" in str(e))

    def test_table_flush_clears_markers_for_that_table(self):
        ModelWithUniques.objects.create(name="One")
        UniqueModel.objects.create(unique_field="One")

        FlushCommand(ModelWithUniques._meta.db_table, default_connection).execute()

        ModelWithUniques.objects.create(name="One")

        with self.assertRaises(IntegrityError):
            UniqueModel.objects.create(unique_field="One")

    def test_recently_deleted_unique_doesnt_come_back(self):
        instance = ModelWithUniques.objects.create(name="One")
        instance.delete()
        self.assertEqual(0, ModelWithUniques.objects.filter(name="One").count())
        self.assertFalse(ModelWithUniques.objects.filter(name="One").exists())
        self.assertFalse(list(ModelWithUniques.objects.all()))  # Triple-check

    def test_conflicting_update_throws_integrity_error(self):
        ModelWithUniques.objects.create(name="One")

        instance = ModelWithUniques.objects.create(name="Two")
        with self.assertRaises(IntegrityError):
            instance.name = "One"
            instance.save()

    def test_delete_works_regardless_of_markers(self):
        stale_instance = ModelWithUniques.objects.create(name="One")

        rpc = transaction._rpc(default_connection.alias)
        # Delete the entity without updating the markers
        key = rpc.key(ModelWithUniques._meta.db_table, stale_instance.pk)
        rpc.delete(key)

        ModelWithUniques.objects.create(name="One")

    def test_unique_combinations_are_returned_correctly(self):
        combos_one = _unique_combinations(ModelWithUniquesOnForeignKey, ignore_pk=True)
        combos_two = _unique_combinations(ModelWithUniquesOnForeignKey, ignore_pk=False)
        rpc = transaction._rpc(default_connection.alias)

        self.assertEqual([['name', 'related_name'], ['name'], ['related_name']], combos_one)
        self.assertEqual([['name', 'related_name'], ['id'], ['name'], ['related_name']], combos_two)

        class Entity(dict):
            def __init__(self, model, id):
                self._key = rpc.key(model, id)

            @property
            def key(self):
                return self._key

        e1 = Entity(ModelWithUniquesOnForeignKey._meta.db_table, 1)
        e1["name"] = "One"
        e1["related_name_id"] = 1

        ids_one = unique_identifiers_from_entity(ModelWithUniquesOnForeignKey, e1, ignore_pk=False)

        self.assertItemsEqual([
            u'tests_modelwithuniquesonforeignkey|id:1',
            u'tests_modelwithuniquesonforeignkey|name:06c2cea18679d64399783748fa367bdd',
            u'tests_modelwithuniquesonforeignkey|related_name_id:1',
            u'tests_modelwithuniquesonforeignkey|name:06c2cea18679d64399783748fa367bdd|related_name_id:1'
        ], ids_one)

    def test_list_field_unique_constaints(self):
        instance1 = UniqueModel.objects.create(unique_field=1, unique_combo_one=1, unique_list_field=["A", "C"])

        with self.assertRaises((IntegrityError, DataError)):
            UniqueModel.objects.create(unique_field=2, unique_combo_one=2, unique_list_field=["A"])

        instance2 = UniqueModel.objects.create(unique_field=2, unique_combo_one=2, unique_list_field=["B"])

        instance2.unique_list_field = instance1.unique_list_field

        with self.assertRaises((IntegrityError, DataError)):
            instance2.save()

        instance1.unique_list_field = []
        instance1.save()

        instance2.save()

    def test_list_field_unique_constraints_validation(self):
        instance1 = UniqueModel(
            unique_set_field={"A"},
            unique_together_list_field=[1],
            unique_field=1,
            unique_combo_one=1,
            unique_list_field=["A", "C"]
        )

        # Without a custom mixin, Django can't construct a unique validation query for a list field
        self.assertRaises(ValueError, instance1.full_clean)
        UniqueModel.__bases__ = (UniquenessMixin,) + UniqueModel.__bases__
        instance1.full_clean()
        instance1.save()

        # Check the uniqueness mixing works with long lists
        instance1.unique_list_field = [x for x in range(31)]
        try:
            instance1.full_clean()
        except NotSupportedError:
            self.fail("Couldn't run unique check on long list field")
            return

        instance2 = UniqueModel(
            unique_set_field={"B"},
            unique_together_list_field=[2],
            unique_field=2,
            unique_combo_one=2,
            unique_list_field=["B", "C"]  # duplicate value C!
        )

        self.assertRaises(ValidationError, instance2.full_clean)
        UniqueModel.__bases__ = (models.Model,)

    def test_set_field_unique_constraints(self):
        instance1 = UniqueModel.objects.create(unique_field=1, unique_combo_one=1, unique_set_field={"A", "C"})

        with self.assertRaises((IntegrityError, DataError)):
            UniqueModel.objects.create(unique_field=2, unique_combo_one=2, unique_set_field={"A"})

        instance2 = UniqueModel.objects.create(unique_field=2, unique_combo_one=2, unique_set_field={"B"})

        instance2.unique_set_field = instance1.unique_set_field

        with self.assertRaises((IntegrityError, DataError)):
            instance2.save()

        instance1.unique_set_field = set()
        instance1.save()

        instance2.save()

        instance2.unique_set_field = set()
        instance2.save()  # You can have two fields with empty sets

    def test_unique_constraints_on_model_with_long_str_pk(self):
        """ Check that an object with a string-based PK of 500 characters (the max that GAE allows)
            can still have unique constraints pointing at it.  (See #242.)
        """
        obj = UniqueModelWithLongPK(pk="x" * 500, unique_field=1)
        obj.save()
        duplicate = UniqueModelWithLongPK(pk="y" * 500, unique_field=1)
        self.assertRaises(IntegrityError, duplicate.save)


class EdgeCaseTests(TestCase):
    def setUp(self):
        super(EdgeCaseTests, self).setUp()

        add_special_index(default_connection, TestUser, "username", IExactIndexer(), "iexact")

        self.u1 = TestUser.objects.create(
            username="A",
            email="test@example.com",
            last_login=datetime.datetime.now().date(),
            id=1, first_name="A", second_name="A"
        )

        self.u2 = TestUser.objects.create(
            username="B",
            email="test@example.com",
            last_login=datetime.datetime.now().date(),
            id=2, first_name="B", second_name="B"
        )

        self.u3 = TestUser.objects.create(
            username="C",
            email="test2@example.com",
            last_login=datetime.datetime.now().date(),
            id=3, first_name="C", second_name="C"
        )

        self.u4 = TestUser.objects.create(
            username="D",
            email="test3@example.com",
            last_login=datetime.datetime.now().date(),
            id=4, first_name="D", second_name="D"
        )

        self.u5 = TestUser.objects.create(
            username="E",
            email="test3@example.com",
            last_login=datetime.datetime.now().date(),
            id=5, first_name="E", second_name="E"
        )

        self.apple = TestFruit.objects.create(name="apple", color="red")
        self.banana = TestFruit.objects.create(name="banana", color="yellow")

    def test_querying_by_date(self):
        instance1 = ModelWithDates.objects.create(start=datetime.date(2014, 1, 1), end=datetime.date(2014, 1, 20))
        instance2 = ModelWithDates.objects.create(start=datetime.date(2014, 2, 1), end=datetime.date(2014, 2, 20))

        self.assertEqual(instance1, ModelWithDates.objects.get(start__lt=datetime.date(2014, 1, 2)))
        self.assertEqual(2, ModelWithDates.objects.filter(start__lt=datetime.date(2015, 1, 1)).count())

        self.assertEqual(instance2, ModelWithDates.objects.get(start__gt=datetime.date(2014, 1, 2)))
        self.assertEqual(instance2, ModelWithDates.objects.get(start__gte=datetime.date(2014, 2, 1)))

    def projection_plus_keys_filtering(self):
        """
            If you do a query like this:

            MyModel.objects.filter(pk__in=[1, 2]).filter(field1="Bananas").values_list("id", "someotherfield")

            Then a projection query is run. The problem is that the entities returned only include
            "id" and "someotherfield"
            but not "field1". Our entity-matches-query code should not run in this situation as we pass
            all filters to the ancestor queries and so any entities returned should match.
        """

        user = TestUser.objects.create(username="test", email="test@example.com")

        self.assertItemsEqual(
            [(user.pk, user.username)],
            TestUser.objects.filter(
                pk__in=[user.pk, user.pk + 1]
            ).filter(
                email="test@example.com"
            ).values_list("id", "username")
        )

    def test_double_starts_with(self):
        qs = TestUser.objects.filter(
            username__startswith='Hello'
        ) | TestUser.objects.filter(username__startswith='Goodbye')

        self.assertEqual(0, qs.count())

        TestUser.objects.create(username="Hello", first_name="A")
        self.assertEqual(1, qs.count())

        TestUser.objects.create(username="Goodbye", first_name="B")
        self.assertEqual(2, qs.count())

        TestUser.objects.create(username="Hello and Goodbye", first_name="C")
        self.assertEqual(3, qs.count())

    def test_impossible_starts_with(self):
        TestUser.objects.create(username="Hello", first_name="A")
        TestUser.objects.create(username="Goodbye", first_name="B")
        TestUser.objects.create(username="Hello and Goodbye", first_name="C")

        qs = TestUser.objects.filter(
            username__startswith='Hello'
        ) & TestUser.objects.filter(username__startswith='Goodbye')
        self.assertEqual(0, qs.count())

    def test_datetime_contains(self):
        """
            Django allows for __contains on datetime field, so that you can search for a specific
            date. This is probably just because SQL allows querying it on a string, and contains just
            turns into a like query. This test just makes sure we behave the same
        """

        instance = DateTimeModel.objects.create()  # Create a DateTimeModel, it has auto_now stuff

        by_datetime = list(DateTimeModel.objects.filter(
                datetime_field__contains=instance.datetime_field.date()
        ))

        # Make sure that if we query a datetime on a date it is properly returned
        self.assertItemsEqual(
            [instance],
            by_datetime
        )

        by_year = list(DateTimeModel.objects.filter(date_field__contains=instance.date_field.year))
        self.assertItemsEqual([instance], by_year)

    # See https://gitlab.com/potato-oss/google-cloud/django-gcloud-connectors/issues/2
    def test_datetime_compare(self):
        from django.utils import timezone
        fake_future = datetime.datetime(2100, 2, 10, 11, 59, 59, tzinfo=timezone.utc)
        fake_now = timezone.datetime(2020, 2, 10, 11, 59, 59)

        instance = DateTimeModel.objects.create(datetime_field=fake_future)

        fetched = DateTimeModel.objects.get(datetime_field__gt=fake_now, pk=instance.pk)

        self.assertEqual(fetched, instance)

    def test_combinations_of_special_indexes(self):
        qs = TestUser.objects.filter(username__iexact='Hello') | TestUser.objects.filter(username__contains='ood')

        self.assertEqual(0, qs.count())

        TestUser.objects.create(username="Hello", first_name="A")
        self.assertEqual(1, qs.count())

        TestUser.objects.create(username="Goodbye", first_name="B")
        self.assertEqual(2, qs.count())

        TestUser.objects.create(username="Hello and Goodbye", first_name="C")
        self.assertEqual(3, qs.count())

    def test_multi_table_inheritance(self):

        parent = MultiTableParent.objects.create(parent_field="parent1")
        child1 = MultiTableChildOne.objects.create(parent_field="child1", child_one_field="child1")
        child2 = MultiTableChildTwo.objects.create(parent_field="child2", child_two_field="child2")

        self.assertEqual(3, MultiTableParent.objects.count())
        self.assertItemsEqual(
            [parent.pk, child1.pk, child2.pk],
            list(MultiTableParent.objects.values_list('pk', flat=True))
        )
        self.assertEqual(1, MultiTableChildOne.objects.count())
        self.assertEqual(child1, MultiTableChildOne.objects.get())

        self.assertEqual(1, MultiTableChildTwo.objects.count())
        self.assertEqual(child2, MultiTableChildTwo.objects.get())

        self.assertEqual(child2, MultiTableChildTwo.objects.get(pk=child2.pk))
        self.assertTrue(MultiTableParent.objects.filter(pk=child2.pk).exists())

    def test_anding_pks(self):
        results = TestUser.objects.filter(id__exact=self.u1.pk).filter(id__exact=self.u2.pk)
        self.assertEqual(list(results), [])

    def test_unusual_queries(self):

        results = list(TestFruit.objects.filter(name__in=["apple", "orange"]))
        self.assertEqual(1, len(results))
        self.assertItemsEqual(["apple"], [x.name for x in results])

        results = TestFruit.objects.filter(name__in=["apple", "banana"])
        self.assertEqual(2, len(results))
        self.assertItemsEqual(["apple", "banana"], [x.name for x in results])

        results = TestFruit.objects.filter(name__in=["apple", "banana"]).values_list('pk', 'color')
        self.assertEqual(2, len(results))
        self.assertItemsEqual([(self.apple.pk, self.apple.color), (self.banana.pk, self.banana.color)], results)

        results = TestUser.objects.all()
        self.assertEqual(5, len(results))

        results = TestUser.objects.filter(username__in=["A", "B"])
        self.assertEqual(2, len(results))
        self.assertItemsEqual(["A", "B"], [x.username for x in results])

        results = TestUser.objects.filter(username__in=["A", "B"]).exclude(username="A")
        self.assertEqual(1, len(results), results)
        self.assertItemsEqual(["B"], [x.username for x in results])

        results = TestUser.objects.filter(username__lt="E")
        self.assertEqual(4, len(results))
        self.assertItemsEqual(["A", "B", "C", "D"], [x.username for x in results])

        results = TestUser.objects.filter(username__lte="E")
        self.assertEqual(5, len(results))

        # Double exclude on different properties not supported
        with self.assertRaises(NotSupportedError):
            # FIXME: This should raise a NotSupportedError, but at the moment it's thrown too late in
            # the process and so Django wraps it as a DataError
            list(TestUser.objects.exclude(username="E").exclude(email="A"))

        results = list(TestUser.objects.exclude(username="E").exclude(username="A"))
        self.assertItemsEqual(["B", "C", "D"], [x.username for x in results])

        results = TestUser.objects.filter(username="A", email="test@example.com")
        self.assertEqual(1, len(results))

        results = TestUser.objects.filter(username__in=["A", "B"]).filter(username__in=["A", "B"])
        self.assertEqual(2, len(results))
        self.assertItemsEqual(["A", "B"], [x.username for x in results])

        results = list(TestUser.objects.filter(username__in=["A", "B"]).filter(username__in=["A"]))
        self.assertEqual(1, len(results))
        self.assertItemsEqual(["A"], [x.username for x in results])

        results = TestUser.objects.filter(pk__in=[self.u1.pk, self.u2.pk]).filter(username__in=["A"])
        self.assertEqual(1, len(results))
        self.assertItemsEqual(["A"], [x.username for x in results])

        results = TestUser.objects.filter(username__in=["A"]).filter(pk__in=[self.u1.pk, self.u2.pk])
        self.assertEqual(1, len(results))
        self.assertItemsEqual(["A"], [x.username for x in results])

        results = list(TestUser.objects.all().exclude(username__in=["A"]))
        self.assertItemsEqual(["B", "C", "D", "E"], [x.username for x in results])

        results = list(TestFruit.objects.filter(name='apple', color__in=[]))
        self.assertItemsEqual([], results)

        results = list(TestUser.objects.all().exclude(username__in=[]))
        self.assertEqual(5, len(results))
        self.assertItemsEqual(["A", "B", "C", "D", "E"], [x.username for x in results])

        results = list(TestUser.objects.all().exclude(username__in=[]).filter(username__in=["A", "B"]))
        self.assertEqual(2, len(results))
        self.assertItemsEqual(["A", "B"], [x.username for x in results])

        results = list(TestUser.objects.all().filter(username__in=["A", "B"]).exclude(username__in=[]))
        self.assertEqual(2, len(results))
        self.assertItemsEqual(["A", "B"], [x.username for x in results])

    def test_empty_string_key(self):
        # Creating
        with self.assertRaises(IntegrityError):
            TestFruit.objects.create(name='')

        # Getting
        with self.assertRaises(TestFruit.DoesNotExist):
            TestFruit.objects.get(name='')

        # Filtering
        results = list(TestFruit.objects.filter(name='').order_by("name"))
        self.assertItemsEqual([], results)

        # Combined filtering
        results = list(TestFruit.objects.filter(name='', color='red').order_by("name"))
        self.assertItemsEqual([], results)

        # IN query
        results = list(TestFruit.objects.filter(name__in=['', 'apple']))
        self.assertItemsEqual([self.apple], results)

    def test_or_queryset(self):
        """
            This constructs an OR query, this is currently broken in the parse_where_and_check_projection
            function. WE MUST FIX THIS!
        """
        q1 = TestUser.objects.filter(username="A")
        q2 = TestUser.objects.filter(username="B")

        self.assertItemsEqual([self.u1, self.u2], list(q1 | q2))

    def test_or_q_objects(self):
        """ Test use of Q objects in filters. """
        query = TestUser.objects.filter(Q(username="A") | Q(username="B"))
        self.assertItemsEqual([self.u1, self.u2], list(query))

    def test_extra_select(self):
        results = TestUser.objects.filter(username='A').extra(select={'is_a': "username = 'A'"})
        self.assertEqual(1, len(results))
        self.assertItemsEqual([True], [x.is_a for x in results])

        results = TestUser.objects.all().exclude(username='A').extra(select={'is_a': "username = 'A'"})
        self.assertEqual(4, len(results))
        self.assertEqual(not any([x.is_a for x in results]), True)

        # Up for debate
        # results = User.objects.all().extra(select={'truthy': 'TRUE'})
        # self.assertEqual(all([x.truthy for x in results]), True)

        results = TestUser.objects.all().extra(select={'truthy': True})
        self.assertEqual(all([x.truthy for x in results]), True)

    def test_counts(self):
        self.assertEqual(5, TestUser.objects.count())
        self.assertEqual(2, TestUser.objects.filter(email="test3@example.com").count())
        self.assertEqual(3, TestUser.objects.exclude(email="test3@example.com").count())
        self.assertEqual(1, TestUser.objects.filter(username="A").exclude(email="test3@example.com").count())
        self.assertEqual(3, TestUser.objects.exclude(username="E").exclude(username="A").count())

        self.assertEqual(3, TestUser.objects.exclude(username__in=["A", "B"]).count())
        self.assertEqual(0, TestUser.objects.filter(email="test@example.com").exclude(username__in=["A", "B"]).count())

    def test_exclude_with__in(self):
        self.assertEqual(
            set([self.u3, self.u4, self.u5]),
            set(list(TestUser.objects.exclude(username__in=["A", "B"])))
        )

    def test_deletion(self):
        count = TestUser.objects.count()
        self.assertTrue(count)

        TestUser.objects.filter(username="A").delete()
        self.assertEqual(count - 1, TestUser.objects.count())

        TestUser.objects.filter(username="B").exclude(username="B").delete()  # Should do nothing
        self.assertEqual(count - 1, TestUser.objects.count())

        TestUser.objects.all().delete()
        count = TestUser.objects.count()
        self.assertFalse(count)

    def test_double_delete(self):
        u1 = TestUser.objects.get(username="A")
        u2 = TestUser.objects.get(username="A")

        u1.delete()
        u2.delete()

    def test_insert_with_existing_key(self):
        user = TestUser.objects.create(id=999, username="test1", last_login=datetime.datetime.now().date())
        self.assertEqual(999, user.pk)

        with self.assertRaises(IntegrityError):
            TestUser.objects.create(id=999, username="test2", last_login=datetime.datetime.now().date())

    def test_included_pks(self):
        ids = [TestUser.objects.get(username="B").pk, TestUser.objects.get(username="A").pk]
        results = TestUser.objects.filter(pk__in=ids).order_by("username")

        self.assertEqual(results[0], self.u1)
        self.assertEqual(results[1], self.u2)

    def test_select_related(self):
        """ select_related should be a no-op... for now """
        user = TestUser.objects.get(username="A")
        Permission.objects.create(user=user, perm="test_perm")
        select_related = [(p.perm, p.user.username) for p in user.permission_set.select_related()]
        self.assertEqual(user.username, select_related[0][1])

    def test_cross_selects(self):
        user = TestUser.objects.get(username="A")
        Permission.objects.create(user=user, perm="test_perm")
        with self.assertRaises(NotSupportedError):
            perms = list(Permission.objects.all().values_list("user__username", "perm"))
            self.assertEqual("A", perms[0][0])

    def test_invalid_id_value_on_insert(self):
        user = TestUser.objects.get(username="A")
        # pass in a User object as if it's an int
        permission = Permission(user_id=user, perm="test_perm")
        with self.assertRaises(TypeError):
            permission.save(force_insert=True)

    def test_values_list_on_pk_does_keys_only_query(self):
        with sleuth.watch("google.cloud.datastore.query.Query.keys_only") as keys_only:
            list(TestUser.objects.all().values_list('pk', flat=True))
            self.assertTrue(keys_only.called)
            self.assertEqual(5, len(TestUser.objects.all().values_list('pk')))

    def test_iexact(self):
        user = TestUser.objects.get(username__iexact="a")
        self.assertEqual("A", user.username)

        add_special_index(default_connection, IntegerModel, "integer_field", IExactIndexer(), "iexact")
        IntegerModel.objects.create(integer_field=1000)
        integer_model = IntegerModel.objects.get(integer_field__iexact=str(1000))
        self.assertEqual(integer_model.integer_field, 1000)

        user = TestUser.objects.get(id__iexact=str(self.u1.id))
        self.assertEqual("A", user.username)

    def test_iexact_containing_underscores(self):
        add_special_index(default_connection, TestUser, "username", IExactIndexer(), "iexact")
        user = TestUser.objects.create(username="A_B", email="test@example.com")
        results = TestUser.objects.filter(username__iexact=user.username.lower())
        self.assertEqual(list(results), [user])

    def test_year(self):
        user = TestUser.objects.create(username="Z", email="z@example.com")
        user.last_login = datetime.datetime(2000, 1, 1, 0, 0, 0)
        user.save()

        self.assertEqual(len(TestUser.objects.filter(last_login__year=3000)), 0)
        self.assertEqual(TestUser.objects.filter(last_login__year=2000).first().pk, user.pk)

    def test_ordering(self):
        users = TestUser.objects.all().order_by("username")

        self.assertEqual(["A", "B", "C", "D", "E"], [x.username for x in users])

        users = TestUser.objects.all().order_by("-username")

        self.assertEqual(["A", "B", "C", "D", "E"][::-1], [x.username for x in users])

        with self.assertRaises(FieldError):
            users = list(TestUser.objects.order_by("bananas"))

        users = TestUser.objects.filter(id__in=[self.u2.id, self.u3.id, self.u4.id]).order_by('id')
        self.assertEqual(["B", "C", "D"], [x.username for x in users])

        users = TestUser.objects.filter(id__in=[self.u2.id, self.u3.id, self.u4.id]).order_by('-id')
        self.assertEqual(["D", "C", "B"], [x.username for x in users])

        users = TestUser.objects.filter(id__in=[self.u1.id, self.u5.id, self.u3.id]).order_by('id')
        self.assertEqual(["A", "C", "E"], [x.username for x in users])

        users = TestUser.objects.filter(id__in=[self.u4.id, self.u5.id, self.u3.id, self.u1.id]).order_by('-id')
        self.assertEqual(["E", "D", "C", "A"], [x.username for x in users])

    def test_dates_query(self):
        z_user = TestUser.objects.create(username="Z", email="z@example.com")
        z_user.last_login = datetime.date(2013, 4, 5)
        z_user.save()

        last_a_login = TestUser.objects.get(username="A").last_login

        dates = TestUser.objects.dates('last_login', 'year')

        self.assertItemsEqual(
            [datetime.date(2013, 1, 1), datetime.date(last_a_login.year, 1, 1)],
            dates
        )

        dates = TestUser.objects.dates('last_login', 'month')
        self.assertItemsEqual(
            [datetime.date(2013, 4, 1), datetime.date(last_a_login.year, last_a_login.month, 1)],
            dates
        )

        dates = TestUser.objects.dates('last_login', 'day')
        self.assertEqual(
            [datetime.date(2013, 4, 5), datetime.date(last_a_login.year, last_a_login.month, last_a_login.day)],
            list(dates)
        )

        dates = TestUser.objects.dates('last_login', 'day', order='DESC')
        self.assertEqual(
            [datetime.date(last_a_login.year, last_a_login.month, last_a_login.day), datetime.date(2013, 4, 5)],
            list(dates)
        )

    @override_settings(DJANGAE_MAX_QUERY_BRANCHES=30)
    def test_in_query(self):
        """ Test that the __in filter works, and that it cannot be used with more than 30 values,
            unless it's used on the PK field.
        """
        # Check that a basic __in query works
        results = list(TestUser.objects.filter(username__in=['A', 'B']))
        self.assertItemsEqual(results, [self.u1, self.u2])
        # Check that it also works on PKs
        results = list(TestUser.objects.filter(pk__in=[self.u1.pk, self.u2.pk]))
        self.assertItemsEqual(results, [self.u1, self.u2])
        # Check that using more than 30 items in an __in query not on the pk causes death
        query = TestUser.objects.filter(username__in=list([x for x in letters[:31]]))
        self.assertRaises(Exception, list, query)
        # Check that it's ok with PKs though
        query = TestUser.objects.filter(pk__in=list(range(1, 32)))
        list(query)
        # Check that it's ok joining filters with pks
        results = list(TestUser.objects.filter(
            pk__in=[self.u1.pk, self.u2.pk, self.u3.pk]).filter(pk__in=[self.u1.pk, self.u2.pk]))
        self.assertItemsEqual(results, [self.u1, self.u2])

    def test_self_relations(self):
        obj = SelfRelatedModel.objects.create()
        obj2 = SelfRelatedModel.objects.create(related=obj)
        self.assertEqual(list(obj.selfrelatedmodel_set.all()), [obj2])

    def test_self_relations_unsaved(self):
        """ Tests self-referencing relations (ForeignKey('self')) """

        obj = SelfRelatedModel()
        SelfRelatedModel(related=obj)
        self.assertEqual(list(obj.selfrelatedmodel_set.all()), [])

    def test_special_indexes_for_empty_fields(self):
        obj = TestFruit.objects.create(name='pear')
        indexes = ['icontains', 'contains', 'iexact', 'iendswith', 'endswith', 'istartswith', 'startswith']
        for index in indexes:
            add_special_index(
                default_connection,
                TestFruit,
                'color',
                get_indexer(TestFruit._meta.get_field("color"), index),
                index
            )
        obj.save()

    def test_special_indexes_for_unusually_long_values(self):
        obj = TestFruit.objects.create(name='pear', color='1234567890-=!@#$%^&*()_+qQWERwertyuiopasdfghjklzxcvbnm')
        indexes = ['icontains', 'contains', 'iexact', 'iendswith', 'endswith', 'istartswith', 'startswith']
        for index in indexes:
            add_special_index(
                default_connection,
                TestFruit,
                'color',
                get_indexer(TestFruit._meta.get_field("color"), index),
                index
            )
        obj.save()

        qry = TestFruit.objects.filter(color__contains='1234567890-=!@#$%^&*()_+qQWERwertyuiopasdfghjklzxcvbnm')
        self.assertEqual(len(list(qry)), 1)
        qry = TestFruit.objects.filter(color__contains='890-=!@#$')
        self.assertEqual(len(list(qry)), 1)
        qry = TestFruit.objects.filter(color__contains='1234567890-=!@#$%^&*()_+qQWERwertyui')
        self.assertEqual(len(list(qry)), 1)
        qry = TestFruit.objects.filter(color__contains='8901')
        self.assertEqual(len(list(qry)), 0)

        qry = TestFruit.objects.filter(color__icontains='1234567890-=!@#$%^&*()_+qQWERWERTYuiopasdfghjklzxcvbnm')
        self.assertEqual(len(list(qry)), 1)
        qry = TestFruit.objects.filter(color__icontains='890-=!@#$')
        self.assertEqual(len(list(qry)), 1)
        qry = TestFruit.objects.filter(color__icontains='1234567890-=!@#$%^&*()_+qQWERwertyuI')
        self.assertEqual(len(list(qry)), 1)
        qry = TestFruit.objects.filter(color__icontains='8901')
        self.assertEqual(len(list(qry)), 0)

    def test_values_list_on_distinct(self):
        TestFruit.objects.create(name="Kiwi", origin="New Zealand", color="Green")
        TestFruit.objects.create(name="Apple", origin="New Zealand", color="Green")
        TestFruit.objects.create(name="Grape", origin="New Zealand", color="Red")
        nz_colours = TestFruit.objects.filter(
            origin="New Zealand"
        ).distinct("color").values_list("color", flat=True)
        self.assertEqual(sorted(nz_colours), ["Green", "Red"])

    def test_empty_key_lookups_work_correctly(self):
        t1 = TestFruit.objects.create(name="Kiwi", origin="New Zealand", color="Green")
        TestFruit.objects.create(name="Apple", origin="New Zealand", color="Green")

        self.assertEqual(
            t1,
            TestFruit.objects.exclude(name="Apple").exclude(name="").get(name="Kiwi")
        )

        self.assertFalse(TestFruit.objects.filter(name="", color="Green"))
        self.assertTrue(TestFruit.objects.filter(Q(name="") | Q(name="Kiwi")).filter(color="Green"))
        self.assertFalse(TestFruit.objects.filter(name="", color__gt="A"))
        self.assertEqual(4, TestFruit.objects.exclude(name="").count())

    def test_additional_indexes_respected(self):
        project, additional = indexing._project_special_indexes.copy(), indexing._app_special_indexes.copy()

        try:
            indexing._project_special_indexes = {}
            indexing._app_special_indexes = {
                TestFruit._meta.db_table: {"name": ["iexact"]}
            }

            t1 = TestFruit.objects.create(name="Kiwi", origin="New Zealand", color="Green")
            self.assertEqual(t1, TestFruit.objects.filter(name__iexact="kiwi").get())
            self.assertFalse(indexing._project_special_indexes)  # Nothing was added
        finally:
            indexing._project_special_indexes = project
            indexing._app_special_indexes = additional


class TestSpecialIndexers(TestCase):

    def setUp(self):
        super(TestSpecialIndexers, self).setUp()

        self.names = [
            'Ola', 'Adam', 'Luke', 'rob', 'Daniel', 'Ela', 'Olga', 'olek',
            'ola', 'Olaaa', 'OlaaA', 'Ola + Ola', '-Test-', '-test-'
        ]
        for name in self.names:
            SpecialIndexesModel.objects.create(name=name)

        self.lists = [
            self.names,
            ['Name', 'name', 'name + name'],
            ['-Tesst-'],
            ['-test-']
        ]
        for i, sample_list in enumerate(self.lists):
            SpecialIndexesModel.objects.create(name=i, sample_list=sample_list)

        self.qry = SpecialIndexesModel.objects.all()

    def test_iexact_lookup(self):
        for name in self.names:
            qry = self.qry.filter(name__iexact=name)
            self.assertEqual(len(qry), len([x for x in self.names if x.lower() == name.lower()]))

    def test_contains_lookup_and_icontains_lookup(self):
        tests = self.names + ['o', 'O', 'la']
        for name in tests:
            qry = self.qry.filter(name__contains=name)
            self.assertEqual(len(qry), len([x for x in self.names if name in x]))

            qry = self.qry.filter(name__icontains=name)
            self.assertEqual(len(qry), len([x for x in self.names if name.lower() in x.lower()]))

    def test_contains_lookup_on_charfield_subclass(self):
        """ Test that the __contains lookup also works on subclasses of the Django CharField, e.g.
            the custom Djangae CharField.
        """
        instance = SpecialIndexesModel.objects.create(name="whatever", nickname="Voldemort")
        query = SpecialIndexesModel.objects.filter(nickname__contains="demo")
        self.assertEqual(list(query), [instance])

    def test_endswith_lookup_and_iendswith_lookup(self):
        tests = self.names + ['a', 'A', 'aa']
        for name in tests:
            qry = self.qry.filter(name__endswith=name)
            self.assertEqual(len(qry), len([x for x in self.names if x.endswith(name)]))

            qry = self.qry.filter(name__iendswith=name)
            self.assertEqual(len(qry), len([x for x in self.names if x.lower().endswith(name.lower())]))

    def test_startswith_lookup_and_istartswith_lookup(self):
        tests = self.names + ['O', 'o', 'ola']
        for name in tests:
            qry = self.qry.filter(name__startswith=name)
            self.assertEqual(len(qry), len([x for x in self.names if x.startswith(name)]))

            qry = self.qry.filter(name__istartswith=name)
            self.assertEqual(len(qry), len([x for x in self.names if x.lower().startswith(name.lower())]))

    def test_regex_lookup_and_iregex_lookup(self):
        tests = [r'([A-Z])\w+', r'([A-Z])\w+\s[+]\s([A-Z])\w+', r'\-Test\-']
        for pattern in tests:
            qry = self.qry.filter(name__regex=pattern)
            self.assertEqual(len(qry), len([x for x in self.names if re.search(pattern, x)]))

            qry = self.qry.filter(name__iregex=pattern)
            self.assertEqual(len(qry), len([x for x in self.names if re.search(pattern, x, flags=re.I)]))

            # Check that the same works for ListField and SetField too
            qry = self.qry.filter(sample_list__item__regex=pattern)
            expected = [
                sample_list for sample_list in self.lists
                if any([bool(re.search(pattern, x)) for x in sample_list])
            ]
            self.assertEqual(len(qry), len(expected))

            qry = self.qry.filter(sample_list__item__iregex=pattern)
            expected = [
                sample_list for sample_list in self.lists
                if any([bool(re.search(pattern, x, flags=re.I)) for x in sample_list])
            ]
            self.assertEqual(len(qry), len(expected))

    def test_item_contains_item_icontains_lookup(self):
        tests = ['O', 'la', 'ola']
        for text in tests:
            qry = self.qry.filter(sample_list__item__contains=text)
            self.assertEqual(len(qry), 1)

            qry = self.qry.filter(sample_list__item__icontains=text)
            self.assertEqual(len(qry), 1)

    def test_item_startswith_item_istartswith_lookup(self):
        tests = ['O', 'ola', 'Ola']
        for text in tests:
            qry = self.qry.filter(sample_list__item__startswith=text)
            self.assertEqual(len(qry), 1)

            qry = self.qry.filter(sample_list__item__istartswith=text)
            self.assertEqual(len(qry), 1)

    def test_item_endswith_item_iendswith_lookup(self):
        tests = ['a', 'la', 'Ola']
        for text in tests:
            qry = self.qry.filter(sample_list__item__endswith=text)
            self.assertEqual(len(qry), 1)

            qry = self.qry.filter(sample_list__item__iendswith=text)
            self.assertEqual(len(qry), 1)


class SliceModel(models.Model):
    field1 = models.CharField(max_length=32)


class SlicingTests(TestCase):

    def test_big_slice(self):
        SliceModel.objects.create(field1="test")
        SliceModel.objects.create(field1="test2")

        self.assertFalse(
            SliceModel.objects.filter(field1__in=["test", "test2"])[9999:]
        )

        self.assertFalse(
            SliceModel.objects.filter(field1__in=["test", "test2"])[9999:10000]
        )

    def test_slicing_multi_query(self):
        SliceModel.objects.create(field1="test")
        SliceModel.objects.create(field1="test2")

        self.assertEqual(
            1,
            len(SliceModel.objects.filter(field1__in=["test", "test2"])[1:])
        )

        self.assertEqual(
            1,
            len(SliceModel.objects.filter(field1__in=["test", "test2"])[:1])
        )

        self.assertEqual(
            2,
            len(SliceModel.objects.filter(field1__in=["test", "test2"])[:2])
        )

        self.assertEqual(
            0,
            len(SliceModel.objects.filter(field1__in=["test", "test2"])[2:])
        )

    def test_slice_params_are_passed_to_query(self):
        for i in range(15):
            SliceModel.objects.create(field1=str(i))

        with sleuth.watch('google.cloud.datastore.query.Query.fetch') as Run:
            qs = SliceModel.objects.order_by("field1")[:5]

            self.assertEqual(5, len(list(qs)))
            self.assertEqual(Run.calls[0].kwargs['limit'], 5)
            self.assertEqual(Run.calls[0].kwargs['offset'], 0)

            qs = SliceModel.objects.order_by("field1")[5:]
            self.assertEqual(10, len(list(qs)))
            self.assertEqual(Run.calls[1].kwargs['limit'], None)
            self.assertEqual(Run.calls[1].kwargs['offset'], 5)

            qs = SliceModel.objects.order_by("field1")[5:10]
            self.assertEqual(5, len(list(qs)))
            self.assertEqual(Run.calls[2].kwargs['limit'], 5)
            self.assertEqual(Run.calls[2].kwargs['offset'], 5)


class NamespaceTests(TestCase):
    multi_db = True
    databases = (
        "default", "nonamespace", "ns1"
    )

    @skipIf("ns1" not in settings.DATABASES, "This test is designed for the Djangae testapp settings")
    def test_database_specific_namespaces(self):
        TestFruit.objects.create(name="Apple", color="Red")
        TestFruit.objects.create(name="Orange", color="Orange")

        TestFruit.objects.using("ns1").create(name="Apple", color="Red")

        self.assertEqual(1, TestFruit.objects.using("ns1").count())
        self.assertEqual(2, TestFruit.objects.count())

        with self.assertRaises(TestFruit.DoesNotExist):
            TestFruit.objects.using("ns1").get(name="Orange")

        try:
            TestFruit.objects.get(name="Orange")
        except TestFruit.DoesNotExist:
            self.fail("Unable to retrieve fruit from the default namespace")

        self.assertEqual(1, TestFruit.objects.filter(name="Orange", color="Orange").count())
        self.assertEqual(0, TestFruit.objects.using("ns1").filter(name="Orange", color="Orange").count())

    def test_no_database_namespace_defaults_to_empty(self):
        """
            Test that creating an object without a namespace makes one that is
            retrievable with just a kind and ID
        """

        TestFruit.objects.using("nonamespace").create(name="Apple", color="Red")
        rpc = transaction._rpc("nonamespace")

        key = rpc.key(TestFruit._meta.db_table, "Apple")
        self.assertTrue(rpc.get(key))

    @skipIf("nonamespace" not in settings.DATABASES, "This test is designed for the Djangae testapp settings")
    def test_move_objects_between_namespaces(self):
        objs = [
            TestFruit.objects.create(name="Banana", color="Black"),
            TestFruit.objects.create(name="Tomato", color="Red"),
        ]
        # First, check that these objects do not exist in the other namespace.
        # We check this in several ways to check that the namespace functionality works in the
        # various different commands of the DB backend
        other_qs = TestFruit.objects.using("nonamespace")
        self.assertEqual(len(other_qs.all()), 0)
        self.assertEqual(other_qs.count(), 0)
        for obj in objs:
            self.assertRaises(TestFruit.DoesNotExist, other_qs.get, name=obj.name)
        # Now re-save both of the objects into the other namespace
        for obj in objs:
            obj.save(using="nonamespace")
        # And now check that they DO exist in that namespace
        self.assertEqual(len(other_qs.all()), 2)
        self.assertEqual(other_qs.count(), 2)
        for obj in objs:
            self.assertEqual(other_qs.get(name=obj.name), obj)
        # Now delete the objects from the original (default) namespace
        TestFruit.objects.all().delete()
        # And now make sure that they exist ONLY in the other namespace
        self.assertEqual(len(TestFruit.objects.all()), 0)
        self.assertEqual(len(other_qs.all()), 2)
        self.assertEqual(TestFruit.objects.count(), 0)
        self.assertEqual(other_qs.count(), 2)
        for obj in objs:
            self.assertRaises(TestFruit.DoesNotExist, TestFruit.objects.get, name=obj.name)
            self.assertEqual(other_qs.get(name=obj.name), obj)


def deferred_func():
    pass


class CascadeDeletionTests(TestCase):
    def test_deleting_less_than_25_items(self):
        zoo = Zoo.objects.create()

        for i in range(10):
            enclosure = Enclosure.objects.create(zoo=zoo)
            for i in range(2):
                Animal.objects.create(enclosure=enclosure)

        self.assertEqual(Animal.objects.count(), 20)

        zoo.delete()

        self.assertEqual(Enclosure.objects.count(), 0)
        self.assertEqual(Animal.objects.count(), 0)

    # see https://github.com/googleapis/google-cloud-python/issues/9921
    @skip("This test should not fail once the emulator bug is fixed")
    def test_deleting_more_than_30_items(self):
        zoo = Zoo.objects.create()

        for i in range(40):
            enclosure = Enclosure.objects.create(zoo=zoo)
            for i in range(2):
                Animal.objects.create(enclosure=enclosure)

        self.assertEqual(Animal.objects.count(), 80)

        zoo.delete()

        self.assertEqual(Enclosure.objects.count(), 0)
        self.assertEqual(Animal.objects.count(), 0)


class AsyncMultiQueryTests(TestCase):

    def test_offset_is_correctly_applied(self):
        TestUser.objects.create(username="Adam", field2="Adam", first_name="A", second_name="A")
        u2 = TestUser.objects.create(username="Bob", first_name="B", second_name="B")
        TestUser.objects.create(username="Chloe", first_name="C", second_name="C")

        qs = TestUser.objects.filter(
            Q(field2="Adam") | Q(username__in=["Adam", "Bob"])
        ).order_by("username")[1:]

        self.assertEqual(len(qs), 1)
        self.assertItemsEqual([u2], qs)

    def test_limit_is_correctly_applied(self):
        TestUser.objects.create(username="Adam", field2="Adam", first_name="A", second_name="A")
        u2 = TestUser.objects.create(username="Bob", first_name="B", second_name="B")
        TestUser.objects.create(username="Chloe", first_name="C", second_name="C")

        qs = TestUser.objects.filter(
            Q(field2="Adam") | Q(username__in=["Adam", "Bob", "Chloe"])
        ).order_by("username")[1:2]

        self.assertEqual(len(qs), 1)
        self.assertItemsEqual([u2], qs)


class UniqueQueryTest(TestCase):
    def test_query_is_unique(self):
        rpc = transaction._rpc(default_connection.alias)
        query = rpc.query(kind="test_testuser")

        # Query with no filter is not unique
        self.assertFalse(query_is_unique(TestUser, query))

        # Query with filter value set to None is not unique
        query.add_filter('username', '=', None)
        self.assertFalse(query_is_unique(TestUser, query))

        # Query on unique field is unique
        query = rpc.query(kind="test_testuser")
        query.add_filter('username', '=', "randy")
        self.assertTrue(query_is_unique(TestUser, query))

        # Query on unique_together fields is unique
        query = rpc.query(kind="test_testuser")
        query.add_filter('first_name', '=', "Randy")
        query.add_filter('second_name', '=', "Munroe")
        self.assertTrue(query_is_unique(TestUser, query))

        # Query on one unique_together field only is not unique
        query = rpc.query(kind="test_testuser")
        query.add_filter('second_name', '=', "Munroe")
        self.assertFalse(query_is_unique(TestUser, query))

        # Query on unique list field is unique only if filtering on one
        # value only, as multiple values are interpreted as "one of"
        query = rpc.query(kind="test_testuniquemodel")
        query.add_filter('unique_list_field', '=', 'a')
        query.add_filter('unique_list_field', '=', 'b')
        self.assertFalse(query_is_unique(UniqueModel, query))

        query = rpc.query(kind="test_testuniquemodel")
        query.add_filter('unique_list_field', '=', 'b')
        self.assertTrue(query_is_unique(UniqueModel, query))

        # Query on pk field is unique
        query = rpc.query(kind="test_testfruit")
        # name is pk
        query.add_filter('__key__', '=', rpc.key('test__testfruit', 'banana'))
        self.assertTrue(query_is_unique(TestFruit, query))

    def test_entity_in_cache(self):
        TestUser.objects.create(username="randy")

        with sleuth.watch("google.cloud.datastore.query.Query.fetch") as fetch:
            list(TestUser.objects.filter(username="randy"))
            self.assertEqual(fetch.call_count, 0)

    def test_entity_not_in_cache(self):
        TestUser.objects.create(username="randy")

        with sleuth.watch("google.cloud.datastore.query.Query.fetch") as fetch:
            with sleuth.fake(
                "gcloudc.db.backends.datastore.meta_queries.caching.get_from_cache",
                return_value=None
            ) as get_from_cache:
                with sleuth.watch(
                    "gcloudc.db.backends.datastore.meta_queries.caching.add_entities_to_cache"
                ) as add_to_cache:
                    list(TestUser.objects.filter(username="randy"))
                    self.assertEqual(add_to_cache.call_count, 1)
                    self.assertEqual(fetch.call_count, 1)
                    self.assertEqual(get_from_cache.call_count, 1)

    def test_keys_only_query_doesnt_cache(self):
        TestUser.objects.create(username="randy")

        with sleuth.watch("gcloudc.db.backends.datastore.meta_queries.caching.add_entities_to_cache") as add_to_cache:
            list(TestUser.objects.filter(username="randy").only("pk"))
            self.assertEqual(add_to_cache.call_count, 0)

    def test_projection_query_doesnt_cache(self):
        TestUser.objects.create(username="randy")

        with sleuth.watch("gcloudc.db.backends.datastore.meta_queries.caching.add_entities_to_cache") as add_to_cache:
            list(TestUser.objects.filter(username="randy").only("first_name"))
            self.assertEqual(add_to_cache.call_count, 0)


class FieldConversionTests(TestCase):
    def test_uuid_field_handled(self):
        instance = UUIDTestModel.objects.create()
        self.assertTrue(isinstance(instance.uuid, uuid.UUID))

        instance.refresh_from_db()
        self.assertTrue(isinstance(instance.uuid, uuid.UUID))


class TransactionFailedErrorTests(TestCase):
    def test_transaction_failed_subclassing_generic_db_exception(self):
        self.assertTrue(issubclass(transaction.TransactionFailedError, DatabaseError))



import copy
import sys

from django.conf import settings

from google.cloud.datastore.key import Key


def key_or_entity_compare(lhs, rhs):
    """
        Returns true if lhs is an entity with a key that matches rhs or
        if lhs is a key which matches rhs, or if lhs is a key which maches
        rhs.key()
    """
    if hasattr(rhs, "key"):
        rhs = rhs.key

    if hasattr(lhs, "key"):
        lhs = lhs.key

    return lhs == rhs


# 8M default cache size. Fairly arbitrary but the lowest instance class (F1) has 128M
# of ram, and can serve 8 Python requests at the same time which gives us 16M per request
# so we use 50% of that by default.
DEFAULT_MAX_CACHE_DICT_SIZE = 1024 * 1024 * 8
MAX_CACHE_DICT_SETTING_NAME = "DJANGAE_CACHE_MAX_CONTEXT_SIZE"


class CacheDict(object):
    """
        This is a special dictionary-like object which does the following:

        1. Copies items in and out to prevent storing references
        2. The cache dict is restricted to a maximum size in bytes
        3. Unaccessed entries are removed first

        The priority of eviction is based on the *value* and not the *keys*. If multiple
        keys point to the same object reference then an access to any of them will mark the
        value as used, if a value is evicted, all the keys pointing to it are removed.
    """

    def __init__(self, max_size_in_bytes=None):
        max_size_in_bytes = max_size_in_bytes or getattr(
            settings, MAX_CACHE_DICT_SETTING_NAME, DEFAULT_MAX_CACHE_DICT_SIZE
        )

        # This is a list of `id(value)` values in priority of most recently used
        # to least recently used
        self.value_priority = []

        # This is a reverse lookup dict of id(value): {key1, key2, ...}
        self.value_references = {}

        # The actual entries, values are normally entities but it makes it easy to
        # understand this class if you think of them as the references they are.
        # Multiple keys can map to the same entity reference
        self._entries = {}

        # THe total size of all values in bytes
        self.total_value_size = 0

        # The max size in bytes that the total values can become
        self.max_size_in_bytes = max_size_in_bytes

    def __deepcopy__(self, memo):
        new_one = CacheDict()
        new_one.update(self)
        return new_one

    def _set_value(self, k, v):
        """
            Sets a value in the _entries dictionary but manages the associated
            data in value_priority and value_references including when a key already
            exists with a different value
        """

        if k in self._entries:
            # We already have a value, we need to clean up
            old_value = self._entries[k]

            # Same object, do nothing
            if id(old_value) == id(v):
                return

            old_key = id(old_value)

            self.value_references[old_key].remove(k)
            del self._entries[k]

            if not self.value_references[old_key]:
                self._purge_value(old_value)

        priority_key = id(v)

        existing_value = priority_key in self.value_priority

        self.value_references.setdefault(priority_key, set()).add(k)
        if priority_key not in self.value_priority:
            self.value_priority.insert(len(self.value_priority) // 2, priority_key)

        self._entries[k] = v

        # If we added a new value to the dict, we increase the used size
        if not existing_value:
            self.total_value_size += sys.getsizeof(v)

    def _check_size_and_limit(self):
        """
            If the dict size is larger than the max specified bytes,
            we remove entities by deleting all their associated keys
        """
        while self.total_value_size > self.max_size_in_bytes:
            next_priority_key = self.value_priority[-1]

            # We intentionally copy the result with list() as this will be manipulated
            # in del self[reference]
            for reference in list(self.value_references[next_priority_key]):
                del self[reference]

    def _set(self, k, v):
        self._set_value(k, v)
        self._check_size_and_limit()

    def set_multi(self, keys, value):
        """
            This is the only public setting API because we don't want to
            duplicate values across keys unnecessarily but we *do* want to copy
            the value passed in by the user to protect against accidental manipulation
            of the cached value
        """

        value = copy.deepcopy(value)  # Copy once
        for k in set(keys):
            # Set the same value for multiple keys
            self._set(k, value)

    def __getitem__(self, k):
        v = self._entries[k]  # Find the entry

        # Move the value up the value priority (remove the id() and add it back at the front)
        priority_key = id(v)
        self.value_priority.remove(priority_key)
        self.value_priority.insert(0, priority_key)
        return copy.deepcopy(v)

    def _purge_value(self, v):
        priority_key = id(v)
        del self.value_references[priority_key]
        self.value_priority.remove(priority_key)
        self.total_value_size -= sys.getsizeof(v)

    def __delitem__(self, k):
        assert set([id(x) for x in self._entries.values()]) == set(self.value_priority)
        v = self._entries[k]
        priority_key = id(v)

        self.value_references[priority_key].remove(k)
        # Only remove from the priority (and adjust the size)
        # if the value no longer exists in the dictionary
        if not self.value_references[priority_key]:
            self._purge_value(v)

        del self._entries[k]

        assert set([id(x) for x in self._entries.values()]) == set(self.value_priority)

    def __repr__(self):
        return "{%s}" % ", ".join([":".join([repr(k), repr(v)]) for k, v in self.items()])

    def __eq__(self, rhs):
        unshared_items = set(self.items()).difference(set(rhs.items()))
        return len(unshared_items) == 0

    def __contains__(self, k):
        return k in self.keys()

    def update(self, other):
        """
            The code here might look weird, but it's the fastest way to do it.

            Go through the other values (remember, we work with reference values)
            and build a dictionary of the id(value): value. This deduplicates values
            which are accessible through multiple keys.

            Then, set_multi the value in self using other.value_references to get the
            list of keys without iterating `other`
        """
        # Find the unique values in the other dictionary
        to_update = {}
        for v in other.values():
            to_update[id(v)] = v

        # Go through, and call set multi (which will perform one copy per value)
        for k, v in to_update.items():
            keys = other.value_references[k]
            self.set_multi(keys, v)

    def __iter__(self):
        return iter(self.keys())

    def get(self, k, default=None):
        try:
            return self[k]
        except KeyError:
            return default

    def keys(self):
        return self._entries.keys()

    def values(self):
        return self._entries.values()

    def items(self):
        """
            Iterate the items, this will not update priorities
        """
        for k in self.keys():
            # Intentionally don't reorganize the key priority if we're iterating
            # that would be *slow* and unlikely to lead to what you want
            yield (k, copy.deepcopy(self._entries[k]))

    def get_reversed(self, value, compare_func=None):
        """
            Returns the keys for the specified value.

            If compare_func is specified then it will be called for
            each value in the cache dict with value until compare_func
            returns True
        """
        results = []
        for k, v in self._entries.items():
            if compare_func and compare_func(v, value):
                results.append(k)
            elif v == value:
                results.append(k)
        return results


class ContextCache(object):
    """ Object via which the stack of Context objects and the settings for the context caching are
        accessed. A separate instance of this should exist per thread.
    """

    def __init__(self):
        self.memcache_enabled = True
        self.context_enabled = True
        self.stack = ContextStack()

    def reset(self, keep_disabled_flags=False):
        from gcloudc.db.transaction import in_atomic_block

        if in_atomic_block():
            raise RuntimeError(
                "Clearing the context cache inside a transaction breaks everything, " "we can't let you do that"
            )

        self.stack = ContextStack()
        if not keep_disabled_flags:
            self.memcache_enabled = True
            self.context_enabled = True


class Context(object):
    def __init__(self, stack):
        self.cache = CacheDict()
        self._stack = stack

    def apply(self, other):
        self.cache.update(other.cache)

        # We have to delete things that don't exist in the other
        keys = list(self.cache)
        for k in keys:
            if k not in other.cache:
                del self.cache[k]

    def cache_entity(self, identifiers, entity, situation):
        assert hasattr(identifiers, "__iter__")

        self.cache.set_multi(identifiers, entity)

    def remove_entity(self, entity_or_key):
        if not isinstance(entity_or_key, Key):
            entity_or_key = entity_or_key.key()

        for identifier in self.cache.get_reversed(entity_or_key, compare_func=key_or_entity_compare):
            del self.cache[identifier]

    def get_entity(self, identifier):
        return self.cache.get(identifier)

    def get_entity_by_key(self, key):
        try:
            identifier = self.cache.get_reversed(key, compare_func=key_or_entity_compare)[0]
        except IndexError:
            return None
        return self.get_entity(identifier)


class ContextStack(object):
    """
        A stack of contexts. This is used to support in-context
        caches for multi level transactions.
    """

    def __init__(self):
        self.stack = [Context(self)]
        self.staged = []

    def push(self):
        self.stack.append(Context(self))  # Empty context

    def pop(self, apply_staged=False, clear_staged=False, discard=False):
        """
            apply_staged: pop normally takes the top of the stack and adds it to a FIFO
            queue. By passing apply_staged it will pop to the FIFO queue then apply the
            queue to the top of the stack.

            clear_staged: pop, and then wipe out any staged contexts.

            discard: Ignores the popped entry in the stack, it's just discarded

            The staged queue will be wiped out if the pop makes the size of the stack one,
            regardless of whether you pass clear_staged or not. This is for safety!
        """
        from . import caching

        if not discard:
            self.staged.insert(0, self.stack.pop())
        else:
            self.stack.pop()

        if apply_staged:
            while self.staged:
                to_apply = self.staged.pop()
                keys = [x.key for x in to_apply.cache.values()]
                if keys:
                    # This assumes that all keys are in the same namespace,
                    # which is almost definitely
                    # going to be the case, but it feels a bit dirty
                    namespace = keys[0].namespace
                    caching.remove_entities_from_cache_by_key(keys, namespace=namespace)

                self.top.apply(to_apply)

        if clear_staged or len(self.stack) == 1:
            self.staged = []

    @property
    def top(self):
        return self.stack[-1]

    @property
    def size(self):
        return len(self.stack)

    @property
    def staged_count(self):
        return len(self.staged)



""" Fake DB API 2.0 for App engine """


class DatabaseError(Exception):
    pass


class IntegrityError(DatabaseError):
    pass


class NotSupportedError(Exception):
    pass


class CouldBeSupportedError(NotSupportedError):
    pass


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class InterfaceError(DatabaseError):
    pass


def Binary(val):
    return val


Error = DatabaseError
Warning = DatabaseError



from contextlib import ContextDecorator

from gcloudc.db.backends.datastore import caching


class DisableCache(ContextDecorator):
    def __enter__(self):
        self.context = caching.get_context()
        self.context.context_enabled = False
        return self

    def __exit__(self, *args, **kwargs):
        self.context.context_enabled = True
        return False


disable_cache = DisableCache



import tempfile

from . import (
    CloudDatastoreRunner,
    locate_command,
)

BaseCommand = locate_command("test")


class Command(CloudDatastoreRunner, BaseCommand):
    USE_MEMORY_DATASTORE_BY_DEFAULT = True

    def _datastore_filename(self):
        print("Creating temporary test database...")

        # This returns the path to a temporary directory which the cloud
        # datastore emulator then intializes as a new
        # datastore
        return tempfile.TemporaryDirectory().name



from . import TestCase
from django.db import models
from gcloudc.db.models.fields.computed import ComputedCharField


class ComputedFieldModel(models.Model):
    def computer(self):
        return "%s_%s" % (self.int_field, self.char_field)

    int_field = models.IntegerField()
    char_field = models.CharField(max_length=50)
    test_field = ComputedCharField(computer, max_length=50)
    method_calc_field = ComputedCharField("computer", max_length=50)

    class Meta:
        app_label = "gcloudc"


# ------------------------


class ComputedFieldTests(TestCase):
    def test_computed_field(self):
        instance = ComputedFieldModel(int_field=1, char_field="test")
        instance.save()
        self.assertEqual(instance.test_field, "1_test")

        # Try getting and saving the instance again
        instance = ComputedFieldModel.objects.get(test_field="1_test")
        instance.save()

    def test_computed_by_method_name_field(self):
        """ Test that a computed field which specifies its "computer" function as a string of
            the name of a method on the model.
        """
        instance = ComputedFieldModel(int_field=2, char_field="test")
        instance.save()
        self.assertEqual(instance.method_calc_field, "2_test")



import sleuth
from django.conf import settings
from django.core.management import call_command

from . import TestCase
from .models import TestUserTwo


class OperationsTests(TestCase):

    def test_bulk_batch_size_override(self):

        with sleuth.watch(
            "gcloudc.db.backends.datastore.commands.InsertCommand.__init__"
        ) as bbs:

            try:
                original = settings.DATABASES["default"].get("OPTIONS", {})
                settings.DATABASES["default"]["OPTIONS"] = {
                    'BULK_BATCH_SIZE': 25
                }

                users = [TestUserTwo(username=str(i)) for i in range(30)]
                TestUserTwo.objects.using("default").bulk_create(users)
            finally:
                settings.DATABASES["default"]["OPTIONS"] = original

            self.assertEqual(bbs.call_count, 2)
            self.assertEqual(len(bbs.calls[0].args[3]), 25)

    def test_flush_large_table(self):
        for i in range(502):
            TestUserTwo.objects.create(username=str(i))

        self.assertEqual(TestUserTwo.objects.count(), 502)
        call_command('flush', interactive=False)
        self.assertFalse(TestUserTwo.objects.exists())



# STANDARD LIB
from abc import ABCMeta
import copy

# THIRD PARTY
from django import forms
from django.db import router, models
from django.db.models.query import QuerySet
from django.db.models.fields.related import ForeignObject, ForeignObjectRel
from django.utils.functional import cached_property
from django.core.exceptions import ImproperlyConfigured, ValidationError
from gcloudc.forms.fields import encode_pk, GenericRelationFormfield, OrderedModelMultipleChoiceField

from django.db.models.query import FlatValuesListIterable

# gcloudc
from gcloudc.core.validators import MinItemsValidator, MaxItemsValidator
from gcloudc.db.models.fields.iterable import IsEmptyLookup, ContainsLookup, OverlapLookup, _serialize_value


class RelatedIteratorRel(ForeignObjectRel):
    def __init__(self, field, to, related_name=None, limit_choices_to=None, on_delete=models.DO_NOTHING, **kwargs):
        self.field = field
        self.model = to
        self.related_name = related_name
        self.related_query_name = None
        self.field_name = None
        self.parent_link = None
        self.on_delete = on_delete
        self.symmetrical = False
        self.limit_choices_to = {} if limit_choices_to is None else limit_choices_to
        self.multiple = True

    def is_hidden(self):
        "Should the related object be hidden?"
        return self.related_name and self.related_name[-1] == "+"

    def set_field_name(self):
        self.field_name = self.field_name or self.model._meta.pk.name

    def get_related_field(self):
        """
        Returns the field in the to' object to which this relationship is tied
        (this is always the primary key on the target model). Provided for
        symmetry with ManyToOneRel.
        """
        return self.model._meta.pk


class OrderedQuerySet(QuerySet):
    def _fetch_all(self):
        """
            Fetch all uses the standard iterator but sorts the values on the
            way out, this maintains the lazy evaluation of querysets
        """

        def locate_pk_column(query):
            pk_name = self.model._meta.pk.name
            if "pk" in query.values_select:
                return query.values_select.index("pk")
            elif pk_name in query.values_select:
                return query.values_select.index(pk_name)

        if self._result_cache is None:
            # Making this work efficiently is tricky. We never want to fetch more than self.ordered_pks
            # and the __getitem__ implementation sets this to a single item, so in that case we only want
            # to fetch one. So we do a few things here:

            # 1. We clone the Queryset as a normal Queryset, then we do a pk__in on the ordered_pks.
            # 2. We add the primary key if this was a values_list query without it being specified, otherwise
            #    we can't match up the ordering
            # 3. We execute the clone, and then remove the additional PK column from the result set

            # There are various combinations to handle depending on whether it's a "flat" values_list or
            # whether or not we added the PK manually to the result set

            clone = QuerySet(model=self.model, query=self.query, using=self._db)
            clone._iterable_class = self._iterable_class
            clone = clone.filter(pk__in=self.ordered_pks)

            pk_col = 0
            pk_added = False

            values_select = clone.query.values_select

            if values_select:
                pk_col = locate_pk_column(clone.query)
                if pk_col is None:
                    # Manually add the PK to the result set
                    clone = clone.values_list(*(["pk"] + list(values_select)))
                    values_select = [x.field.name for x in clone.query.select]
                    pk_col = 0
                    pk_added = True

            # Hit the database
            results = list(clone)

            ordered_results = []
            pk_hash = {}

            flat = self._iterable_class == FlatValuesListIterable

            for x in results:
                if isinstance(x, models.Model):
                    # standard query case
                    pk_hash[x.pk] = x
                elif len(values_select) == 1:
                    # Only PK case
                    pk_hash[x if flat else x[pk_col]] = x
                else:
                    # Multiple columns (either passed in, or as a result of the PK being added)
                    if flat:
                        pk_hash[x[pk_col]] = x[1] if pk_added else x
                    else:
                        pk_hash[x[pk_col]] = x[1:] if pk_added else x

            for pk in self.ordered_pks:
                obj = pk_hash.get(pk)
                if obj:
                    ordered_results.append(obj)
            self._result_cache = ordered_results
        if self._prefetch_related_lookups and not self._prefetch_done:
            self._prefetch_related_objects()

    def _clone(self, *args, **kwargs):
        """
            We need to attach the ordered_pk list on the clone to it continues
            through the chain
        """
        c = super(OrderedQuerySet, self)._clone(*args, **kwargs)
        c.ordered_pks = self.ordered_pks[:]
        return c

    def __getitem__(self, k):
        """
            Ok to get query slicing working properly we can't allow it
            to set limits on the internal query, because that is not the
            behavior we want, instead we slice the internal ordered_pk list
        """
        if not isinstance(k, (slice,) + (int,)):
            raise TypeError

        assert (not isinstance(k, slice) and (k >= 0)) or (
            isinstance(k, slice) and (k.start is None or k.start >= 0) and (k.stop is None or k.stop >= 0)
        )

        if self._result_cache is not None:
            return self._result_cache[k]

        if isinstance(k, slice):
            qs = self._clone()
            if k.start is not None:
                start = int(k.start)
            else:
                start = None
            if k.stop is not None:
                stop = int(k.stop)
            else:
                stop = None
            qs.ordered_pks = qs.ordered_pks[start:stop]
            return list(qs)[:: k.step] if k.step else qs

        qs = self._clone()
        qs.ordered_pks = [qs.ordered_pks[k]]
        return list(qs)[0]


class RelatedIteratorManagerBase(object):
    def __init__(self, model, field, instance, reverse):
        super(RelatedIteratorManagerBase, self).__init__()
        self.model = model
        self.instance = instance
        self.field = field
        self.reverse = reverse
        self.ordered = field.retain_underlying_order

        if reverse:
            self.core_filters = {"%s__contains" % self.field.column: instance.pk}
        else:
            self.core_filters = {"pk__in": field.value_from_object(instance)}

    def get_prefetch_queryset(self, instances, queryset=None):
        related_model = self.field.related_model if hasattr(self.field, "related_model") else self.field.related.to
        if not queryset:
            queryset = related_model.objects.all()

        matchers = {}
        related_ids = set()
        for instance in instances:
            for related_id in getattr(instance, self.field.attname):
                related_ids.add(related_id)
                matchers.setdefault(related_id, []).append(instance.pk)

        queryset = queryset.filter(pk__in=related_ids)

        def duplicator(queryset):
            """
                We have to duplicate the related objects for each source instance that
                was passed in so we can convince Django to do the right thing when merging
            """
            for related_thing in queryset:
                for i, matcher in enumerate(matchers[related_thing.pk]):
                    if i == 0:
                        # Optimisation to prevent unnecessary extra copy
                        ret = related_thing
                    else:
                        ret = copy.copy(related_thing)

                    # Set an id so we can fetch things out in the lambdas below
                    ret._prefetch_instance_id = matcher
                    yield ret

        return (
            duplicator(queryset),
            lambda inst: inst._prefetch_instance_id,
            lambda obj: obj.pk,
            False,
            self.field.name,  # Use the field name as the cache name
            True
        )

    def get_queryset(self):
        db = self._db or router.db_for_read(self.model, instance=self.instance)

        if (
            hasattr(self.instance, "_prefetched_objects_cache")
            and self.field.name in self.instance._prefetched_objects_cache
        ):

            qs = self.instance._prefetched_objects_cache[self.field.name]

            if self.ordered:
                lookup = {}
                for item in qs._result_cache:
                    lookup[item.pk] = item

                # If this is a ListField make sure the result set retains the right order
                qs._result_cache = []
                for pk in getattr(self.instance, self.field.attname):
                    qs._result_cache.append(lookup[pk])
                return qs
            else:
                return qs
        elif self.ordered and not self.reverse:
            values = self.field.value_from_object(self.instance)
            qcls = OrderedQuerySet(self.model, using=db)
            qcls.ordered_pks = values[:]
        else:
            qcls = super(RelatedIteratorManagerBase, self).get_queryset()
        return qcls.using(db)._next_is_sticky().filter(**self.core_filters)

    def add(self, *values):
        for value in values:
            if not isinstance(value, self.model):
                raise TypeError("'%s' instance expected, got %r" % (self.model._meta.object_name, value))

            if not value.pk:
                raise ValueError("Model instances must be saved before they can be added to a related set")

            field_value = self.field.value_from_object(self.instance)
            # Depending on the type of iterable, but we want to maintain the
            # related .add() behavior
            if isinstance(field_value, list):
                field_value.append(value.pk)
            elif isinstance(field_value, set):
                field_value.add(value.pk)

    def remove(self, value):
        field_value = self.field.value_from_object(self.instance)
        # Depending on the type of iterable, but we want to maintain the
        # related .remove() behavior
        if isinstance(field_value, list):
            field_value.remove(value.pk)
        elif isinstance(field_value, set):
            field_value.discard(value.pk)

    def clear(self):
        setattr(self.instance, self.field.attname, self.field.default())

    def __len__(self):
        return len(self.field.value_from_object(self.instance))


def create_related_iter_manager(superclass, rel):
    """ Create a manager for the (reverse) relation which subclasses the related model's default manager. """

    class RelatedIteratorManager(RelatedIteratorManagerBase, superclass):
        pass

    return RelatedIteratorManager


class RelatedIteratorObjectsDescriptor(object):
    # This class provides the functionality that makes the related-object
    # managers available as attributes on a model class, for fields that have
    # multiple "remote" values and have a ManyToManyField pointed at them by
    # some other model (rather than having a ManyToManyField themselves).
    # In the example "publication.article_set", the article_set attribute is a
    # ManyRelatedObjectsDescriptor instance.
    def __init__(self, related):
        self.related = related  # RelatedObject instance

    @cached_property
    def related_manager_cls(self):
        """
            Dynamically create a class that subclasses the related
            model's default manager.
        """

        manager = self.related.related_model._default_manager.__class__

        return create_related_iter_manager(manager, self.related.field.remote_field)

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self

        rel_model = self.related.related_model
        rel_field = self.related.field

        manager = self.related_manager_cls(model=rel_model, field=rel_field, instance=instance, reverse=True)

        return manager

    def __set__(self, obj, value):
        raise AttributeError("You can't set the reverse relation directly")


class ReverseRelatedObjectsDescriptor(object):
    # This class provides the functionality that makes the related-object
    # managers available as attributes on a model class, for fields that have
    # multiple "remote" values and have a ManyToManyField defined in their
    # model (rather than having another model pointed *at* them).
    # In the example "article.publications", the publications attribute is a
    # ReverseManyRelatedObjectsDescriptor instance.
    def __init__(self, m2m_field):
        self.field = m2m_field

    @cached_property
    def related_manager_cls(self):
        # Dynamically create a class that subclasses the related model's
        # default manager.
        return create_related_iter_manager(
            self.field.remote_field.model._default_manager.__class__, self.field.remote_field.model
        )

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self

        manager = self.related_manager_cls(
            model=self.field.remote_field.model, field=self.field, instance=instance, reverse=False
        )

        return manager

    def __set__(self, obj, value):
        obj.__dict__[self.field.attname] = self.field.to_python([x.pk for x in value])


class RelatedContainsLookup(ContainsLookup):
    def get_prep_lookup(self):
        # If this is an instance, then we want to lookup based on the PK
        if isinstance(self.rhs, models.Model):
            self.rhs = self.rhs.pk
        return super(RelatedContainsLookup, self).get_prep_lookup()


class RelatedOverlapLookup(OverlapLookup):
    def get_prep_lookup(self):
        # Transform any instances to primary keys
        self.rhs = [x.pk if isinstance(x, models.Model) else x for x in self.rhs]
        return super(RelatedOverlapLookup, self).get_prep_lookup()


class RelatedIteratorField(ForeignObject):

    __metaclass__ = ABCMeta

    requires_unique_target = False
    generate_reverse_relation = True
    empty_strings_allowed = False
    retain_underlying_order = False

    rel_class = RelatedIteratorRel

    def __init__(self, to, limit_choices_to=None, related_name=None, on_delete=models.DO_NOTHING, **kwargs):
        # Make sure that we do nothing on cascade by default
        self._check_sane_on_delete_value(on_delete)

        from_fields = ["self"]
        to_fields = [None]

        min_length, max_length = self._sanitize_min_and_max_length(kwargs)

        super(RelatedIteratorField, self).__init__(
            to,
            on_delete,
            from_fields,
            to_fields,
            related_name=related_name,
            limit_choices_to=limit_choices_to,
            **kwargs
        )

        # Now that self.validators has been set up, we can add the min/max legnth validators
        if min_length is not None:
            self.validators.append(MinItemsValidator(min_length))
        if max_length is not None:
            self.validators.append(MaxItemsValidator(max_length))

    def _check_sane_on_delete_value(self, on_delete):
        if on_delete == models.CASCADE:
            raise ImproperlyConfigured(
                "on_delete=CASCADE is disabled for iterable fields as this will "
                "wipe out the instance when the field still has related objects"
            )

        if on_delete in (models.SET_NULL, models.SET_DEFAULT):
            raise ImproperlyConfigured(
                "Using an on_delete value of {} will cause undesirable behavior"
                " (e.g. wipeout the entire list) if you really want to do that "
                "then use models.SET instead and return an empty list/set"
            )

    def _sanitize_min_and_max_length(self, kwargs):
        # Pop the 'min_length' and 'max_length' from the kwargs, if they're there, as this avoids
        # 'min_length' causing an error when calling super()
        min_length = kwargs.pop("min_length", None)
        max_length = kwargs.pop("max_length", None)

        # Check that if there's a min_length that blank is not True.  This is partly because it
        # doesn't make sense, and partly because if the value (i.e. the list or set) is empty then
        # Django will skip the validators, thereby skipping the min_length check.
        if min_length and kwargs.get("blank"):
            raise ImproperlyConfigured("Setting blank=True and min_length=%d is contradictory." % min_length)

        return min_length, max_length

    def _get_lookup_constraint(self, constraint_class, alias, targets, sources, lookups, raw_value):
        from django.core import exceptions
        from django.db.models.sql.where import AND

        root_constraint = constraint_class()

        # we need some checks from ForeignObject.get_lookup_constraint
        # before we could check the lookups.
        assert len(targets) == len(sources)
        if len(lookups) > 1:
            raise exceptions.FieldError("%s does not support nested lookups" % self.__class__.__name__)

        lookup_type = lookups[0]
        target = targets[0]
        source = sources[0]

        # use custom Lookups when applicable
        if lookup_type in ["isempty", "overlap", "contains"]:
            if lookup_type == "isempty":
                root_constraint.add(IsEmptyLookup(target.get_col(alias, source), raw_value), AND)
            elif lookup_type == "overlap":
                root_constraint.add(RelatedOverlapLookup(target.get_col(alias, source), raw_value), AND)
            elif lookup_type == "contains":
                root_constraint.add(RelatedContainsLookup(target.get_col(alias, source), raw_value), AND)
            return root_constraint
        elif lookup_type in ("in", "exact", "isnull"):
            raise TypeError(
                "%s doesn't allow exact, in or isnull. Use contains, overlap or isempty respectively"
                % self.__class__.__name__
            )

        return super(RelatedIteratorField, self).get_lookup_constraint(
            constraint_class, alias, targets, sources, lookups, raw_value
        )

    def deconstruct(self):
        name, path, args, kwargs = super(RelatedIteratorField, self).deconstruct()
        # We hardcode a number of arguments for RelatedIteratorField, those arguments need to be removed here
        for hardcoded_kwarg in ["to_fields", "from_fields", "on_delete"]:
            del kwargs[hardcoded_kwarg]

        return name, path, args, kwargs

    def get_attname(self):
        return "%s_ids" % self.name

    def get_attname_column(self):
        attname = self.get_attname()
        column = self.db_column or attname
        return attname, column

    def contribute_to_class(self, cls, name):
        # To support multiple relations to self, it's useful to have a non-None
        # related name on symmetrical relations for internal reasons. The
        # concept doesn't make a lot of sense externally ("you want me to
        # specify *what* on my non-reversible relation?!"), so we set it up
        # automatically. The funky name reduces the chance of an accidental
        # clash.
        rel_to = self.remote_field.model if self.remote_field else None
        if rel_to == "self" or rel_to == cls._meta.object_name:
            self.remote_field.related_name = "%s_rel_+" % name

        super(RelatedIteratorField, self).contribute_to_class(cls, name)

        # Add the descriptor for the m2m relation
        setattr(cls, self.name, ReverseRelatedObjectsDescriptor(self))

    def contribute_to_related_class(self, cls, related):
        # Internal M2Ms (i.e., those with a related name ending with '+')
        # and swapped models don't get a related descriptor.
        if not self.remote_field.is_hidden() and not related.model._meta.swapped:
            setattr(cls, related.get_accessor_name(), RelatedIteratorObjectsDescriptor(related))

    def get_db_prep_save(self, *args, **kwargs):
        ret = super(RelatedIteratorField, self).get_db_prep_save(*args, **kwargs)

        if not ret:
            return None

        if isinstance(ret, set):
            ret = list(ret)

        ret = [self.remote_field.model._meta.pk.clean(x, self.remote_field.model) for x in ret]

        return ret

    def get_db_prep_lookup(self, *args, **kwargs):
        ret = super(RelatedIteratorField, self).get_db_prep_lookup(*args, **kwargs)

        if not ret:
            return []

        if isinstance(ret, set):
            ret = list(ret)
        return ret

    def value_to_string(self, obj):
        return u"[" + ",".join(_serialize_value(o) for o in self.value_from_object(obj)) + "]"

    def formfield(self, **kwargs):
        db = kwargs.pop("using", None)
        form_class = kwargs.pop("form_class", forms.ModelMultipleChoiceField)
        defaults = {
            "form_class": form_class,
            "queryset": self.remote_field.model._default_manager.using(db).complex_filter(
                self.remote_field.limit_choices_to
            ),
        }
        defaults.update(kwargs)
        # If initial is passed in, it's a list of related objects, but the
        # MultipleChoiceField takes a list of IDs.
        if defaults.get("initial") is not None:
            initial = defaults["initial"]
            if callable(initial):
                initial = initial()
            defaults["initial"] = [i._get_pk_val() for i in initial]
        return super(RelatedIteratorField, self).formfield(**defaults)

    def get_lookup(self, lookup_name):
        if lookup_name == "isempty":
            return IsEmptyLookup
        elif lookup_name == "overlap":
            return RelatedOverlapLookup
        elif lookup_name == "contains":
            return RelatedContainsLookup
        elif lookup_name in ("in", "exact", "isnull"):
            raise TypeError("RelatedIteratorFields don't allow exact, in or isnull. Use contains, overlap or isempty")

        return super(RelatedIteratorField, self).get_lookup(lookup_name)

    def to_python(self, value):
        if value is None:
            return list()

        # Deal with deserialization from a string
        if isinstance(value, str):
            if not (value.startswith("[") and value.endswith("]")):
                raise ValidationError("Invalid input for {} instance", type(self).__name__)

            value = value[1:-1].strip()

            if not value:
                return list()

            ids = [
                self.remote_field.model._meta.pk.to_python(x.strip("'").strip('"'))
                for x in value.split(",")
                if len(value) > 2
            ]
            # Annoyingly Django special cases FK and M2M in the Python deserialization code,
            # to assign to the attname, whereas all other fields (including this one) are required to
            # populate field.name instead. So we have to query here... we have no choice :(
            objs = self.remote_field.model._default_manager.db_manager("default").in_bulk(ids)

            # retain order
            return [objs.get(_id) for _id in ids if _id in objs]

        return list(value)


RelatedIteratorField.register_lookup(RelatedContainsLookup)
RelatedIteratorField.register_lookup(RelatedOverlapLookup)
RelatedIteratorField.register_lookup(IsEmptyLookup)


class RelatedSetField(RelatedIteratorField):
    def db_type(self, connection):
        return "set"

    def __init__(self, *args, **kwargs):

        kwargs["default"] = set
        kwargs["null"] = True

        super(RelatedSetField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(RelatedSetField, self).deconstruct()
        # We hardcode a number of arguments for RelatedIteratorField, those arguments need to be removed here
        for hardcoded_kwarg in ["default", "null"]:
            del kwargs[hardcoded_kwarg]

        return name, path, args, kwargs

    def to_python(self, value):
        return set(super(RelatedSetField, self).to_python(value))

    def save_form_data(self, instance, data):
        setattr(instance, self.attname, set())  # Wipe out existing things
        for value in data:
            # If this is a model instance then add the pk
            if isinstance(value, self.remote_field.model):
                value = value.pk

            field = getattr(instance, self.attname)
            field.add(value)


class RelatedListField(RelatedIteratorField):
    retain_underlying_order = True

    def db_type(self, connection):
        return "list"

    def __init__(self, *args, **kwargs):

        kwargs["default"] = list
        kwargs["null"] = True
        self.remove_duplicates = kwargs.pop("remove_duplicates", False)

        super(RelatedListField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(RelatedListField, self).deconstruct()
        # We hardcode a number of arguments for RelatedIteratorField, those arguments need to be removed here
        for hardcoded_kwarg in ["default", "null"]:
            del kwargs[hardcoded_kwarg]

        kwargs["remove_duplicates"] = self.remove_duplicates

        return name, path, args, kwargs

    def formfield(self, **kwargs):
        """
        Specify a OrderedModelMultipleChoiceField for the `form_class` so we
        can retain ordering.
        """
        # change the form_class in defaults from using ModelMultipleChoiceField
        # in preference for the gcloudc subclass OrderedModelMultipleChoiceField
        kwargs["form_class"] = kwargs.pop("form_class", OrderedModelMultipleChoiceField)
        return super(RelatedListField, self).formfield(**kwargs)

    def save_form_data(self, instance, data):
        setattr(instance, self.attname, [])  # Wipe out existing things
        for value in data:
            # If this is a model instance then grab the PK
            if isinstance(value, self.remote_field.model):
                value = value.pk

            # Add to the underlying list
            field = getattr(instance, self.attname)
            field.append(value)

    def pre_save(self, model_instance, add):
        value = super(RelatedListField, self).pre_save(model_instance, add)

        if value and self.remove_duplicates:
            seen = set()
            value = [x for x in value if not (x in seen or seen.add(x))]
            # We should also update the model attribute to hold correct reference
            setattr(model_instance, self.attname, value)

        return value


class GRCreator(property):

    """
    This is like django.db.models.fields.subclassing.Creator, but does its
    magic on get instead of set.
    """

    def __init__(self, field):
        self.field = field
        self.attname = field.get_attname()
        super(GRCreator, self).__init__(self.get, self.set)

    def get(self, obj):
        cache_attr = "_{}_cache".format(self.attname)
        if getattr(obj, cache_attr, None) is None:
            setattr(obj, cache_attr, self.field.__get__(obj))
        return getattr(obj, cache_attr)

    def set(self, obj, value):
        obj.__dict__[self.attname] = self.field.get_prep_value(value)
        if isinstance(value, models.Model):
            setattr(obj, "_{}_cache".format(self.attname), value)


class GRReverseCreator(property):

    """Prevent forms setting whatever_id to an object"""

    def __init__(self, field):
        self.field = field
        self.attname = field.get_attname()
        super(GRReverseCreator, self).__init__(self.get, self.set)

    def get(self, obj):
        return obj.__dict__[self.attname]

    def set(self, obj, value):
        if not isinstance(value, str):
            value = self.field.get_prep_value(value)
        obj.__dict__[self.attname] = value


class GenericRelationField(models.Field):
    """ Similar to django.contrib.contenttypes.generic.GenericForeignKey, but
        doesn't require all of the contenttypes cruft.

        Empty NULL values are specifically encoded so they can also be unique, this is different
        from Django's FK behaviour.
    """

    description = "Generic one-way relation field"
    form_class = GenericRelationFormfield

    def __init__(self, *args, **kwargs):
        kwargs.update(max_length=255)

        if "db_index" not in kwargs:
            kwargs["db_index"] = True

        super(GenericRelationField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        super(GenericRelationField, self).contribute_to_class(cls, name)
        setattr(cls, self.name, GRCreator(self))
        setattr(cls, self.attname, GRReverseCreator(self))

    def get_attname(self):
        return "{}_id".format(self.name)

    def set_attributes_from_name(self, name):
        super(GenericRelationField, self).set_attributes_from_name(name)

    @classmethod
    def encode_pk(cls, pk, obj_cls):
        return encode_pk(pk, obj_cls)

    def get_internal_type(self):
        return "CharField"

    def get_prep_value(cls, value):
        return cls.form_class.to_string(value)

    def __get__(self, instance):
        if instance is None:
            return self

        value = getattr(instance, self.attname)

        from gcloudc.forms.fields import decode_pk
        from gcloudc.db.backends.datastore.utils import get_model_from_db_table

        if value is None:
            return None

        model_ref, pk = decode_pk(value)
        try:
            return get_model_from_db_table(model_ref).objects.get(pk=pk)
        except AttributeError:
            raise ImproperlyConfigured("Unable to find model with db_table: {}".format(model_ref))

    def formfield(self, **kwargs):
        defaults = {"form_class": self.form_class}
        defaults.update(kwargs)
        return super(GenericRelationField, self).formfield(**defaults)

    def get_validator_unique_lookup_type(self):
        raise NotImplementedError()



import copy
import threading
from functools import cmp_to_key, partial
from itertools import groupby

from django.conf import settings
from google.cloud.datastore.key import Key

from . import POLYMODEL_CLASS_ATTRIBUTE, caching
from .query_utils import compare_keys, get_filter, is_keys_only
from .utils import django_ordering_comparison, entity_matches_query


class AsyncMultiQuery(object):
    """
        Runs multiple queries simultaneously and merges the result sets based on the
        shared ordering.
    """

    # Testing seems to show that more threads == better, but I'm concerned if we
    # raise this too high we'll start hitting bottlenecks elsewhere. Serious performance
    # testing needs to happen. Theoretically I think we could make THREAD_COUNT == len(queries)
    # but I'd rather prove that doesn't cause problems before I do it!
    THREAD_COUNT = 8

    def __init__(self, queries, orderings):
        self._queries = [copy.copy(x) for x in queries]
        self._orderings = orderings
        self._min_max_cache = {}

        # When set, this is called on the query before .Run() is called
        # Which allows you to manipulate the options. Recommend this is set/unset
        # in a try/finally
        self._query_decorator = None
        self._keys_only = False

    def keys_only(self):
        self._keys_only = True
        for query in self._queries:
            query.keys_only()

    def _spawn_thread(self, i, query, result_queues, **query_run_args):
        """
            Spawns a thread to return a queries resultset

            *Note* by evaluating the entire query results in the thread we ruin the datastore
            query batching in the situation that you:

             a. Have limited the query
             b. Have a large number of results in one or more branches of the OR

            Basically, if you do this:

            MyModel.objects.filter(field1__in=("A", "B"))[:1000]

            and you have 1000 results with "A" and 1000 results with "B" all
            2000 results will be fetched even though you asked for 1000. However, this is
            not the most likely situation for a MultiQuery when normally few results will be returned
            by each branch. Threading seems to help in the common case but we can revisit
            when we have more data. If threading isn't worth the cost we can revert to just using
            async queries like Google's multiquery does.
        """

        keys_only = self._keys_only

        class Thread(threading.Thread):
            def __init__(self, query, *args, **kwargs):
                self.query = query
                self.results_fetched = False
                super(Thread, self).__init__(*args, **kwargs)

            def run(self):
                # Evaluate the result set in the thread, but return an iterator
                # so we can change this if necessary without breaking assumptions elsewhere
                result_queues[i] = (x.key if keys_only else x for x in self.query.fetch(**query_run_args))
                self.results_fetched = True

        if self._query_decorator:
            query = self._query_decorator(query)

        thread = Thread(query)
        thread.start()
        return thread

    def _fetch_results(self, limit=None):
        """
            Returns a list of generators (one for each query in the multi query)
            which generate entity results (or keys if it's keys_only)

            Uses multiple threads to submit RPC calls
        """

        threads = []

        # We need to grab a set of results per query
        result_queues = [None] * len(self._queries)

        # Go through the queries, trigger new threads as they become available
        for i, query in enumerate(self._queries):

            # Iterate while we have a full thread list
            while len(threads) >= self.THREAD_COUNT:
                try:
                    complete = next(x for x in threads if x.results_fetched)
                except StopIteration:
                    # No threads available, continue waiting
                    continue

                # Remove the complete thread
                complete.join()
                threads.remove(complete)

            # Spawn a new thread
            threads.append(self._spawn_thread(i, query, result_queues, limit=limit))

        [x.join() for x in threads]  # Wait until all the threads are done

        return result_queues

    def _compare_entities(self, lhs, rhs):
        def cmp(a, b):
            # Treat None as less than everything for the purpose
            # of ordering
            if a is None and b is None:
                return 0
            elif a is None:
                return -1
            elif b is None:
                return 1

            return (a > b) - (a < b)

        if isinstance(lhs, Key) and isinstance(rhs, Key):
            return compare_keys(lhs, rhs)

        def get_extreme_if_list_property(entity_key, column, value, descending):
            if not isinstance(value, list):
                return value

            if (entity_key, column) in self._min_max_cache:
                return self._min_max_cache[(entity_key, column)]

            if descending:
                value = min(value)
            else:
                value = max(value)
            self._min_max_cache[(entity_key, column)] = value

        if not lhs:
            return cmp(lhs, rhs)

        for column in self._orderings:
            descending = column.startswith("-")
            column = column.lstrip("-")

            lhs_value = lhs.key if column == "__key__" else lhs[column]
            rhs_value = rhs.key if column == "__key__" else rhs[column]

            lhs_value = get_extreme_if_list_property(lhs.key, column, lhs_value, descending)
            rhs_value = get_extreme_if_list_property(lhs.key, column, rhs_value, descending)

            if isinstance(lhs_value, Key) and isinstance(rhs_value, Key):
                result = compare_keys(lhs_value, rhs_value)
            else:
                result = cmp(lhs_value, rhs_value)

            if descending:
                result = -result
            if result:
                return result

        return compare_keys(lhs.key, rhs.key)

    def fetch(self, offset=None, limit=None):
        """
            Returns an iterator through the result set.

            This calls _fetch_results which returns a list of iterators,
            where each is the result of a single query. This function does a
            zig-zag merge of the entities. It starts by creating a list of the next
            entry in each resultset, then iteratively picks the next entity and then
            fills the slot from the counterpart result set until all the slots are None.
        """
        self._min_max_cache = []

        # We have to assume that one branch might return all the results and as
        # offsetting is done by skipping results we need to get offset + limit results
        # from each branch
        results = self._fetch_results(limit=(offset or 0) + limit if limit is not None else None)

        # Go through each outstanding result queue and store
        # the next entry of each (None if the result queue is done)
        next_entries = [None] * len(results)
        for i, queue in enumerate(results):
            try:
                next_entries[i] = next(results[i])
            except StopIteration:
                next_entries[i] = None

        returned_count = 0
        yielded_count = 0

        seen_keys = set()  # For de-duping results
        while any(next_entries):

            def get_next():
                idx, lowest = None, None

                for i, entry in enumerate(next_entries):
                    if entry is None:
                        continue

                    if lowest is None or self._compare_entities(entry, lowest) < 0:
                        idx, lowest = i, entry

                # Move the queue along if we found the entry there
                if lowest is not None:
                    try:
                        next_entries[idx] = next(results[idx])
                    except StopIteration:
                        next_entries[idx] = None

                return lowest

            # Find the next entry from the available queues
            next_entity = get_next()

            # No more entries if this is the case
            if next_entity is None:
                break

            next_key = next_entity if isinstance(next_entity, Key) else next_entity.key

            # Make sure we haven't seen this result before before yielding
            if next_key not in seen_keys:
                returned_count += 1
                seen_keys.add(next_key)

                if offset and returned_count <= offset:
                    # We haven't hit the offset yet, so just
                    # keep fetching entities
                    continue

                yielded_count += 1
                yield next_entity

                if limit and yielded_count == limit:
                    break


def _convert_entity_based_on_query_options(entity, keys_only, projection):
    if keys_only:
        return entity.key

    if projection:
        keys = list(entity.keys())
        for k in keys:
            if k not in list(projection) + [POLYMODEL_CLASS_ATTRIBUTE]:
                del entity[k]

    return entity


# The max number of entities in a resultset that will be cached
# if a query returns more than this number then only the first ones
# will be cached
DEFAULT_MAX_ENTITY_COUNT = 8


class QueryByKeys(object):
    """ Does the most efficient fetching possible for when we have the keys of the entities we want. """

    def __init__(self, connection, model, queries, ordering, namespace):
        # Imported here for potential circular import and isolation reasons
        from .dnf import DEFAULT_MAX_ALLOWABLE_QUERIES

        # `queries` should be filtered by __key__ with keys that have the namespace applied to them.
        # `namespace` is passed for explicit niceness (mostly so that we don't have to assume that
        # all the keys belong to the same namespace, even though they will).
        def _get_key(query):
            result = get_filter(query, ("__key__", "="))
            return result

        def compare_queries(lhs, rhs):
            return compare_keys(_get_key(lhs), _get_key(rhs))

        self.connection = connection
        self.model = model
        self.namespace = namespace

        # groupby requires that the iterable is sorted by the given key before grouping
        self.queries = sorted(queries, key=cmp_to_key(compare_queries))
        self.query_count = len(self.queries)
        self.queries_by_key = {a: list(b) for a, b in groupby(self.queries, _get_key)}

        self.max_allowable_queries = getattr(settings, "DJANGAE_MAX_QUERY_BRANCHES", DEFAULT_MAX_ALLOWABLE_QUERIES)
        self.can_multi_query = self.query_count < self.max_allowable_queries

        self.ordering = ordering
        self.kind = queries[0].kind
        self._keys_only_override = False

    def keys_only(self):
        self._keys_only_override = True

    def fetch(self, limit=None, offset=None):
        """
            Here are the options:

            1. Single key, hit memcache
            2. Multikey projection, async MultiQueries with ancestors chained
            3. Full select, datastore get
        """
        from gcloudc.db.backends.datastore import transaction
        from gcloudc.db.backends.datastore.caching import MAX_CACHE_COUNT

        base_query = self.queries[0]
        key_count = len(self.queries_by_key)

        is_projection = False

        cache_results = True
        results = None

        if key_count == 1:
            # FIXME: Potentially could use get_multi in memcache and the make a query
            # for whatever remains
            key = next(iter(self.queries_by_key))
            assert(isinstance(key, Key))

            result = caching.get_from_cache_by_key(key)
            if result is not None:
                results = [result]
                cache_results = False  # Don't update cache, we just got it from there

        client = transaction._rpc(self.connection)
        if results is None:
            if base_query.projection and self.can_multi_query:
                is_projection = True

                # If we can multi-query in a single query, we do so using a number of
                # ancestor queries (to stay consistent) otherwise, we just do a
                # datastore Get, but this will return extra data over the RPC
                to_fetch = (offset or 0) + limit if limit else None

                # strip the optional "-" prefix from order (indicating descending order)
                additional_cols = set([
                    x if x[0] != "-" else x[1:]
                    for x in self.ordering
                    if x not in base_query.projection
                ])

                multi_query = []
                orderings = base_query.order
                for key, queries in self.queries_by_key.items():
                    for query in queries:
                        if additional_cols:
                            # We need to include additional orderings in the projection so that we can
                            # sort them in memory. Annoyingly that means reinstantiating the queries
                            query = client.query(
                                kind=query.kind,
                                filters=query.filters,
                                projection=list(base_query.projection) + list(additional_cols),
                                namespace=self.namespace,
                            )

                        query.ancestor = key  # Make this an ancestor query
                        multi_query.append(query)

                if len(multi_query) == 1:
                    results = multi_query[0].fetch(limit=to_fetch)
                else:
                    results = AsyncMultiQuery(multi_query, orderings).fetch(limit=to_fetch)
            else:
                # Can pass 1000 keys to a datastore.Get, so if there are more we need to do them
                # in multiple gets
                MAX_ALLOWED_GET = 1000

                all_keys = list(self.queries_by_key.keys())
                next_keys, remaining_keys = all_keys[:MAX_ALLOWED_GET], all_keys[MAX_ALLOWED_GET:]

                results = []
                while next_keys:
                    results.extend(client.get(next_keys))
                    next_keys, remaining_keys = remaining_keys[:MAX_ALLOWED_GET], remaining_keys[MAX_ALLOWED_GET:]

        def iter_results(results):
            returned = 0

            # This is safe, because Django is fetching all results any way :(
            sorted_results = sorted(results, key=cmp_to_key(partial(django_ordering_comparison, self.ordering)))
            sorted_results = [result for result in sorted_results if result is not None]

            if cache_results and sorted_results:
                caching.add_entities_to_cache(
                    self.model,
                    sorted_results[:MAX_CACHE_COUNT],
                    caching.CachingSituation.DATASTORE_GET,
                    self.namespace,
                )

            for result in sorted_results:
                if is_projection:
                    matches_query = True
                else:
                    matches_query = any(entity_matches_query(result, qry) for qry in self.queries_by_key[result.key])

                if not matches_query:
                    continue

                if offset and returned < offset:
                    # Skip entities based on offset
                    returned += 1
                    continue
                else:
                    yield _convert_entity_based_on_query_options(
                        result, self._keys_only_override or is_keys_only(base_query), base_query.projection
                    )

                    returned += 1

                    # If there is a limit, we might be done!
                    if limit is not None and returned == (offset or 0) + limit:
                        break

        return iter_results(results)


class NoOpQuery(object):
    def fetch(self, limit, offset):
        return []


class UniqueQuery(object):
    """
        This mimics a normal query but hits the cache if possible. It must
        be passed the set of unique fields that form a unique constraint
    """

    def __init__(self, unique_identifier, gae_query, model, namespace):
        self._identifier = unique_identifier
        self._gae_query = gae_query
        self._model = model
        self._namespace = namespace
        self._keys_only = False

        self.kind = gae_query.kind

    def keys_only(self):
        self._gae_query.keys_only()
        self._keys_only = True

    def fetch(self, limit, offset):
        do_cache = True
        ret = caching.get_from_cache(self._identifier, self._namespace)

        if ret is not None:
            if not entity_matches_query(ret, self._gae_query):
                ret = None
            else:
                ret = [ret]

        if ret is None:
            ret = list(self._gae_query.fetch(limit=limit, offset=offset))
        else:
            # We don't want to cache a result already coming from the cache...
            do_cache = False

        if self._keys_only or self._gae_query.projection:
            # ...nor if it's coming from a keys_only/projection query
            do_cache = False

        if len(ret) == 1 and do_cache:
            caching.add_entities_to_cache(
                self._model,
                [ret[0]],
                caching.CachingSituation.DATASTORE_GET,
                self._namespace,
            )

        return iter(ret)
